import sys

import numpy as np

try:
    from IPython.lib.pretty import pprint
except ImportError:
    pprint = None
import pytest

from brian2 import Equations, Expression, Hz, Unit, farad, metre, ms, mV, second, volt
from brian2.core.namespace import DEFAULT_UNITS
from brian2.equations.equations import (
    BOOLEAN,
    DIFFERENTIAL_EQUATION,
    FLOAT,
    INTEGER,
    PARAMETER,
    SUBEXPRESSION,
    EquationError,
    SingleEquation,
    check_identifier_basic,
    check_identifier_constants,
    check_identifier_functions,
    check_identifier_reserved,
    check_identifier_units,
    dimensions_and_type_from_string,
    extract_constant_subexpressions,
    parse_string_equations,
)
from brian2.equations.refractory import check_identifier_refractory
from brian2.groups.group import Group
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    DimensionMismatchError,
    get_dimensions,
)


# a simple Group for testing
class SimpleGroup(Group):
    def __init__(self, variables, namespace=None):
        self.variables = variables
        self.namespace = namespace


@pytest.mark.codegen_independent
def test_utility_functions():
    unit_namespace = DEFAULT_UNITS

    # Some simple tests whether the namespace returned by
    # get_default_namespace() makes sense
    assert "volt" in unit_namespace
    assert "ms" in unit_namespace
    assert unit_namespace["ms"] is ms
    assert unit_namespace["ms"] is unit_namespace["msecond"]
    for unit in unit_namespace.values():
        assert isinstance(unit, Unit)

    assert dimensions_and_type_from_string("second") == (second.dim, FLOAT)
    assert dimensions_and_type_from_string("1") == (DIMENSIONLESS, FLOAT)
    assert dimensions_and_type_from_string("volt") == (volt.dim, FLOAT)
    assert dimensions_and_type_from_string("second ** -1") == (Hz.dim, FLOAT)
    assert dimensions_and_type_from_string("farad / metre**2") == (
        (farad / metre**2).dim,
        FLOAT,
    )
    assert dimensions_and_type_from_string("boolean") == (DIMENSIONLESS, BOOLEAN)
    assert dimensions_and_type_from_string("integer") == (DIMENSIONLESS, INTEGER)
    with pytest.raises(ValueError):
        dimensions_and_type_from_string("metr / second")
    with pytest.raises(ValueError):
        dimensions_and_type_from_string("metre **")
    with pytest.raises(ValueError):
        dimensions_and_type_from_string("5")
    with pytest.raises(ValueError):
        dimensions_and_type_from_string("2 / second")
    # Only the use of base units is allowed
    with pytest.raises(ValueError):
        dimensions_and_type_from_string("farad / cm**2")


@pytest.mark.codegen_independent
def test_identifier_checks():
    legal_identifiers = ["v", "Vm", "V", "x", "ge", "g_i", "a2", "gaba_123"]
    illegal_identifiers = ["_v", "1v", "ü", "ge!", "v.x", "for", "else", "if"]

    for identifier in legal_identifiers:
        try:
            check_identifier_basic(identifier)
            check_identifier_reserved(identifier)
        except ValueError as ex:
            raise AssertionError(
                f'check complained about identifier "{identifier}": {ex}'
            )

    for identifier in illegal_identifiers:
        with pytest.raises(SyntaxError):
            check_identifier_basic(identifier)

    for identifier in ("t", "dt", "xi", "i", "N"):
        with pytest.raises(SyntaxError):
            check_identifier_reserved(identifier)

    for identifier in ("not_refractory", "refractory", "refractory_until"):
        with pytest.raises(SyntaxError):
            check_identifier_refractory(identifier)

    for identifier in ("exp", "sin", "sqrt"):
        with pytest.raises(SyntaxError):
            check_identifier_functions(identifier)

    for identifier in ("e", "pi", "inf"):
        with pytest.raises(SyntaxError):
            check_identifier_constants(identifier)

    for identifier in ("volt", "second", "mV", "nA"):
        with pytest.raises(SyntaxError):
            check_identifier_units(identifier)

    # Check identifier registry
    assert check_identifier_basic in Equations.identifier_checks
    assert check_identifier_reserved in Equations.identifier_checks
    assert check_identifier_refractory in Equations.identifier_checks
    assert check_identifier_functions in Equations.identifier_checks
    assert check_identifier_constants in Equations.identifier_checks
    assert check_identifier_units in Equations.identifier_checks

    # Set up a dummy identifier check that disallows the variable name
    # gaba_123 (that is otherwise valid)
    def disallow_gaba_123(identifier):
        if identifier == "gaba_123":
            raise SyntaxError("I do not like this name")

    Equations.check_identifier("gaba_123")
    old_checks = set(Equations.identifier_checks)
    Equations.register_identifier_check(disallow_gaba_123)
    with pytest.raises(SyntaxError):
        Equations.check_identifier("gaba_123")
    Equations.identifier_checks = old_checks

    # registering a non-function should not work
    with pytest.raises(ValueError):
        Equations.register_identifier_check("no function")


@pytest.mark.codegen_independent
def test_parse_equations():
    """Test the parsing of equation strings"""
    # A simple equation
    eqs = parse_string_equations("dv/dt = -v / tau : 1")
    assert len(eqs) == 1 and "v" in eqs and eqs["v"].type == DIFFERENTIAL_EQUATION
    assert eqs["v"].dim is DIMENSIONLESS

    # A complex one
    eqs = parse_string_equations(
        """
        dv/dt = -(v +
                  ge + # excitatory conductance
                  I # external current
                 )/ tau : volt
        dge/dt = -ge / tau_ge : volt
        I = sin(2 * pi * f * t) : volt
        f : Hz (constant)
        b : boolean
        n : integer
        """
    )
    assert len(eqs) == 6
    assert "v" in eqs and eqs["v"].type == DIFFERENTIAL_EQUATION
    assert "ge" in eqs and eqs["ge"].type == DIFFERENTIAL_EQUATION
    assert "I" in eqs and eqs["I"].type == SUBEXPRESSION
    assert "f" in eqs and eqs["f"].type == PARAMETER
    assert "b" in eqs and eqs["b"].type == PARAMETER
    assert "n" in eqs and eqs["n"].type == PARAMETER
    assert eqs["f"].var_type == FLOAT
    assert eqs["b"].var_type == BOOLEAN
    assert eqs["n"].var_type == INTEGER
    assert eqs["v"].dim is volt.dim
    assert eqs["ge"].dim is volt.dim
    assert eqs["I"].dim is volt.dim
    assert eqs["f"].dim is Hz.dim
    assert eqs["v"].flags == []
    assert eqs["ge"].flags == []
    assert eqs["I"].flags == []
    assert eqs["f"].flags == ["constant"]

    duplicate_eqs = """
    dv/dt = -v / tau : 1
    v = 2 * t : 1
    """
    with pytest.raises(EquationError):
        parse_string_equations(duplicate_eqs)
    parse_error_eqs = [
        """
        dv/d = -v / tau : 1
        x = 2 * t : 1
        """,
        """
        dv/dt = -v / tau : 1 : volt
        x = 2 * t : 1
        """,
        "dv/dt = -v / tau : 2 * volt",
        "dv/dt = v / second : boolean",
    ]
    for error_eqs in parse_error_eqs:
        with pytest.raises((ValueError, EquationError, TypeError)):
            parse_string_equations(error_eqs)


@pytest.mark.codegen_independent
def test_correct_replacements():
    """Test replacing variables via keyword arguments"""
    # replace a variable name with a new name
    eqs = Equations("dv/dt = -v / tau : 1", v="V")
    # Correct left hand side
    assert ("V" in eqs) and not ("v" in eqs)
    # Correct right hand side
    assert ("V" in eqs["V"].identifiers) and not ("v" in eqs["V"].identifiers)

    # replace a variable name with a value
    eqs = Equations("dv/dt = -v / tau : 1", tau=10 * ms)
    assert not "tau" in eqs["v"].identifiers


@pytest.mark.codegen_independent
def test_wrong_replacements():
    """Tests for replacements that should not work"""
    # Replacing a variable name with an illegal new name
    with pytest.raises(SyntaxError):
        Equations("dv/dt = -v / tau : 1", v="illegal name")
    with pytest.raises(SyntaxError):
        Equations("dv/dt = -v / tau : 1", v="_reserved")
    with pytest.raises(SyntaxError):
        Equations("dv/dt = -v / tau : 1", v="t")

    # Replacing a variable name with a value that already exists
    with pytest.raises(EquationError):
        Equations(
            """
            dv/dt = -v / tau : 1
            dx/dt = -x / tau : 1
            """,
            v="x",
        )

    # Replacing a model variable name with a value
    with pytest.raises(ValueError):
        Equations("dv/dt = -v / tau : 1", v=3 * mV)

    # Replacing with an illegal value
    with pytest.raises(SyntaxError):
        Equations("dv/dt = -v/tau : 1", tau=np.arange(5))


@pytest.mark.codegen_independent
def test_substitute():
    # Check that Equations.substitute returns an independent copy
    eqs = Equations("dx/dt = x : 1")
    eqs2 = eqs.substitute(x="y")

    # First equation should be unaffected
    assert len(eqs) == 1 and "x" in eqs
    assert eqs["x"].expr == Expression("x")

    # Second equation should have x substituted by y
    assert len(eqs2) == 1 and "y" in eqs2
    assert eqs2["y"].expr == Expression("y")


@pytest.mark.codegen_independent
def test_construction_errors():
    """
    Test that the Equations constructor raises errors correctly
    """
    # parse error
    with pytest.raises(EquationError):
        Equations("dv/dt = -v / tau volt")
    with pytest.raises(EquationError):
        Equations("dv/dt = -v / tau : volt second")

    # incorrect unit definition
    with pytest.raises(EquationError):
        Equations("dv/dt = -v / tau : mvolt")
    with pytest.raises(EquationError):
        Equations("dv/dt = -v / tau : voltage")
    with pytest.raises(EquationError):
        Equations("dv/dt = -v / tau : 1.0*volt")

    # Only a single string or a list of SingleEquation objects is allowed
    with pytest.raises(TypeError):
        Equations(None)
    with pytest.raises(TypeError):
        Equations(42)
    with pytest.raises(TypeError):
        Equations(["dv/dt = -v / tau : volt"])

    # duplicate variable names
    with pytest.raises(EquationError):
        Equations(
            """
            dv/dt = -v / tau : volt
            v = 2 * t/second * volt : volt
            """
        )

    eqs = [
        SingleEquation(
            DIFFERENTIAL_EQUATION, "v", volt.dim, expr=Expression("-v / tau")
        ),
        SingleEquation(
            SUBEXPRESSION, "v", volt.dim, expr=Expression("2 * t/second * volt")
        ),
    ]
    with pytest.raises(EquationError):
        Equations(eqs)

    # illegal variable names
    with pytest.raises(SyntaxError):
        Equations("ddt/dt = -dt / tau : volt")
    with pytest.raises(SyntaxError):
        Equations("dt/dt = -t / tau : volt")
    with pytest.raises(SyntaxError):
        Equations("dxi/dt = -xi / tau : volt")
    with pytest.raises(SyntaxError):
        Equations("for : volt")
    with pytest.raises((EquationError, SyntaxError)):
        Equations("d1a/dt = -1a / tau : volt")
    with pytest.raises(SyntaxError):
        Equations("d_x/dt = -_x / tau : volt")

    # xi in a subexpression
    with pytest.raises(EquationError):
        Equations(
            """
            dv/dt = -(v + I) / (5 * ms) : volt
            I = second**-1*xi**-2*volt : volt
            """
        )

    # more than one xi
    with pytest.raises(EquationError):
        Equations(
            """
            dv/dt = -v / tau + xi/tau**.5 : volt
            dx/dt = -x / tau + 2*xi/tau : volt
            tau : second
            """
        )
    # using not-allowed flags
    eqs = Equations("dv/dt = -v / (5 * ms) : volt (flag)")
    eqs.check_flags({DIFFERENTIAL_EQUATION: ["flag"]})  # allow this flag
    with pytest.raises(ValueError):
        eqs.check_flags({DIFFERENTIAL_EQUATION: []})
    with pytest.raises(ValueError):
        eqs.check_flags({})
    with pytest.raises(ValueError):
        eqs.check_flags({SUBEXPRESSION: ["flag"]})
    with pytest.raises(ValueError):
        eqs.check_flags({DIFFERENTIAL_EQUATION: ["otherflag"]})
    eqs = Equations("dv/dt = -v / (5 * ms) : volt (flag1, flag2)")
    eqs.check_flags({DIFFERENTIAL_EQUATION: ["flag1", "flag2"]})  # allow both flags
    # Don't allow the two flags in combination
    with pytest.raises(ValueError):
        eqs.check_flags(
            {DIFFERENTIAL_EQUATION: ["flag1", "flag2"]},
            incompatible_flags=[("flag1", "flag2")],
        )
    eqs = Equations(
        """
        dv/dt = -v / (5 * ms) : volt (flag1)
        dw/dt = -w / (5 * ms) : volt (flag2)
        """
    )
    # They should be allowed when used independently
    eqs.check_flags(
        {DIFFERENTIAL_EQUATION: ["flag1", "flag2"]},
        incompatible_flags=[("flag1", "flag2")],
    )

    # Circular subexpression
    with pytest.raises(ValueError):
        Equations(
            """
            dv/dt = -(v + w) / (10 * ms) : 1
            w = 2 * x : 1
            x = 3 * w : 1
            """
        )

    # Boolean/integer differential equations
    with pytest.raises(TypeError):
        Equations("dv/dt = -v / (10*ms) : boolean")
    with pytest.raises(TypeError):
        Equations("dv/dt = -v / (10*ms) : integer")


@pytest.mark.codegen_independent
def test_unit_checking():
    # dummy Variable class
    class S:
        def __init__(self, dimensions):
            self.dim = get_dimensions(dimensions)

    # inconsistent unit for a differential equation
    eqs = Equations("dv/dt = -v : volt")
    group = SimpleGroup({"v": S(volt)})
    with pytest.raises(DimensionMismatchError):
        eqs.check_units(group, {})

    eqs = Equations("dv/dt = -v / tau: volt")
    group = SimpleGroup(namespace={"tau": 5 * mV}, variables={"v": S(volt)})
    with pytest.raises(DimensionMismatchError):
        eqs.check_units(group, {})
    group = SimpleGroup(namespace={"I": 3 * second}, variables={"v": S(volt)})
    eqs = Equations("dv/dt = -(v + I) / (5 * ms): volt")
    with pytest.raises(DimensionMismatchError):
        eqs.check_units(group, {})

    eqs = Equations(
        """
        dv/dt = -(v + I) / (5 * ms): volt
        I : second
        """
    )
    group = SimpleGroup(variables={"v": S(volt), "I": S(second)}, namespace={})
    with pytest.raises(DimensionMismatchError):
        eqs.check_units(group, {})

    # inconsistent unit for a subexpression
    eqs = Equations(
        """
        dv/dt = -v / (5 * ms) : volt
        I = 2 * v : amp
        """
    )
    group = SimpleGroup(variables={"v": S(volt), "I": S(second)}, namespace={})
    with pytest.raises(DimensionMismatchError):
        eqs.check_units(group, {})


@pytest.mark.codegen_independent
def test_properties():
    """
    Test accessing the various properties of equation objects
    """
    tau = 10 * ms
    eqs = Equations(
        """
        dv/dt = -(v + I)/ tau : volt
        I = sin(2 * 22/7. * f * t)* volt : volt
        f = freq * Hz: Hz
        freq : 1
        """
    )
    assert (
        len(eqs.diff_eq_expressions) == 1
        and eqs.diff_eq_expressions[0][0] == "v"
        and isinstance(eqs.diff_eq_expressions[0][1], Expression)
    )
    assert eqs.diff_eq_names == {"v"}
    assert (
        len(eqs.eq_expressions) == 3
        and {name for name, _ in eqs.eq_expressions} == {"v", "I", "f"}
        and all((isinstance(expr, Expression) for _, expr in eqs.eq_expressions))
    )
    assert len(eqs.eq_names) == 3 and eqs.eq_names == {"v", "I", "f"}
    assert set(eqs.keys()) == {"v", "I", "f", "freq"}
    # test that the equations object is iterable itself
    assert all(isinstance(eq, SingleEquation) for eq in eqs.values())
    assert all(isinstance(eq, str) for eq in eqs)
    assert (
        len(eqs.ordered) == 4
        and all(isinstance(eq, SingleEquation) for eq in eqs.ordered)
        and [eq.varname for eq in eqs.ordered] == ["f", "I", "v", "freq"]
    )
    assert [eq.unit for eq in eqs.ordered] == [Hz, volt, volt, 1]
    assert eqs.names == {"v", "I", "f", "freq"}
    assert eqs.parameter_names == {"freq"}
    assert eqs.subexpr_names == {"I", "f"}
    dimensions = eqs.dimensions
    assert set(dimensions.keys()) == {"v", "I", "f", "freq"}
    assert dimensions["v"] is volt.dim
    assert dimensions["I"] is volt.dim
    assert dimensions["f"] is Hz.dim
    assert dimensions["freq"] is DIMENSIONLESS
    assert eqs.names == set(eqs.dimensions.keys())
    assert eqs.identifiers == {"tau", "volt", "Hz", "sin", "t"}

    # stochastic equations
    assert len(eqs.stochastic_variables) == 0
    assert eqs.stochastic_type is None

    eqs = Equations("""dv/dt = -v / tau + 0.1*second**-.5*xi : 1""")
    assert eqs.stochastic_variables == {"xi"}
    assert eqs.stochastic_type == "additive"

    eqs = Equations(
        "dv/dt = -v / tau + 0.1*second**-.5*xi_1 +  0.1*second**-.5*xi_2: 1"
    )
    assert eqs.stochastic_variables == {"xi_1", "xi_2"}
    assert eqs.stochastic_type == "additive"

    eqs = Equations("dv/dt = -v / tau + 0.1*second**-1.5*xi*t : 1")
    assert eqs.stochastic_type == "multiplicative"

    eqs = Equations("dv/dt = -v / tau + 0.1*second**-1.5*xi*v : 1")
    assert eqs.stochastic_type == "multiplicative"


@pytest.mark.codegen_independent
def test_concatenation():
    eqs1 = Equations(
        """
        dv/dt = -(v + I) / tau : volt
        I = sin(2*pi*freq*t) : volt
        freq : Hz
        """
    )

    # Concatenate two equation objects
    eqs2 = Equations("dv/dt = -(v + I) / tau : volt") + Equations(
        """
        I = sin(2*pi*freq*t) : volt
        freq : Hz
        """
    )

    # Concatenate using "in-place" addition (which is not actually in-place)
    eqs3 = Equations("dv/dt = -(v + I) / tau : volt")
    eqs3 += Equations(
        """
        I = sin(2*pi*freq*t) : volt
        freq : Hz
        """
    )

    # Concatenate with a string (will be parsed first)
    eqs4 = Equations("dv/dt = -(v + I) / tau : volt")
    eqs4 += """I = sin(2*pi*freq*t) : volt
               freq : Hz"""

    # Concatenating with something that is not a string should not work
    with pytest.raises(TypeError):
        eqs4 + 5

    # The string representation is canonical, therefore it should be identical
    # in all cases
    assert str(eqs1) == str(eqs2)
    assert str(eqs2) == str(eqs3)
    assert str(eqs3) == str(eqs4)


@pytest.mark.codegen_independent
def test_extract_subexpressions():
    eqs = Equations(
        """
        dv/dt = -v / (10*ms) : 1
        s1 = 2*v : 1
        s2 = -v : 1 (constant over dt)
        """
    )
    variable, constant = extract_constant_subexpressions(eqs)
    assert [var in variable for var in ["v", "s1", "s2"]]
    assert variable["s1"].type == SUBEXPRESSION
    assert variable["s2"].type == PARAMETER
    assert constant["s2"].type == SUBEXPRESSION


@pytest.mark.codegen_independent
def test_repeated_construction():
    eqs1 = Equations("dx/dt = x : 1")
    eqs2 = Equations("dx/dt = x : 1", x="y")
    assert len(eqs1) == 1
    assert "x" in eqs1
    assert eqs1["x"].expr == Expression("x")
    assert len(eqs2) == 1
    assert "y" in eqs2
    assert eqs2["y"].expr == Expression("y")


@pytest.mark.codegen_independent
def test_str_repr():
    """
    Test the string representation (only that it does not throw errors).
    """
    tau = 10 * ms
    eqs = Equations(
        """
        dv/dt = -(v + I)/ tau : volt (unless refractory)
        I = sin(2 * 22/7. * f * t)* volt : volt
        f : Hz
        """
    )
    assert len(str(eqs)) > 0
    assert len(repr(eqs)) > 0

    # Test str and repr of SingleEquations explicitly (might already have been
    # called by Equations
    for eq in eqs.values():
        assert (len(str(eq))) > 0
        assert (len(repr(eq))) > 0


@pytest.mark.codegen_independent
def test_dependency_calculation():
    eqs = Equations(
        """
        dv/dt = I_m / C_m : volt
        I_m = I_ext + I_pas : amp
        I_ext = 1*nA + sin(2*pi*100*Hz*t)*nA : amp
        I_pas = g_L*(E_L - v) : amp
        """
    )
    deps = eqs.dependencies
    assert set(deps.keys()) == {"v", "I_m", "I_ext", "I_pas"}

    # v depends directly on I_m, on I_ext and I_pas via I_m, and on v via I_m -> I_pas
    assert len(deps["v"]) == 4
    assert {d.equation.varname for d in deps["v"]} == {"I_m", "I_ext", "I_pas", "v"}
    expected_via = {
        "I_m": (),
        "I_pas": ("I_m",),
        "I_ext": ("I_m",),
        "v": ("I_m", "I_pas"),
    }
    assert all([d.via == expected_via[d.equation.varname] for d in deps["v"]])

    # I_m depends directly on I_ext and I_pas, and on v via I_pas
    assert len(deps["I_m"]) == 3
    assert {d.equation.varname for d in deps["I_m"]} == {"I_ext", "I_pas", "v"}
    expected_via = {"I_ext": (), "I_pas": (), "v": ("I_pas",)}
    assert all([d.via == expected_via[d.equation.varname] for d in deps["I_m"]])

    # I_ext does not depend on anything
    assert len(deps["I_ext"]) == 0

    # I_pas depends on v directly
    assert len(deps["I_pas"]) == 1
    assert deps["I_pas"][0].equation.varname == "v"
    assert deps["I_pas"][0].via == ()


@pytest.mark.codegen_independent
@pytest.mark.skipif(pprint is None, reason="ipython is not installed")
def test_ipython_pprint():
    from io import StringIO

    eqs = Equations(
        """
        dv/dt = -(v + I)/ tau : volt (unless refractory)
        I = sin(2 * 22/7. * f * t)* volt : volt
        f : Hz
        """
    )
    # Test ipython's pretty printing
    old_stdout = sys.stdout
    string_output = StringIO()
    sys.stdout = string_output
    pprint(eqs)
    assert len(string_output.getvalue()) > 0
    sys.stdout = old_stdout


if __name__ == "__main__":
    test_utility_functions()
    test_identifier_checks()
    test_parse_equations()
    test_correct_replacements()
    test_substitute()
    test_wrong_replacements()
    test_construction_errors()
    test_concatenation()
    test_unit_checking()
    test_properties()
    test_extract_subexpressions()
    test_repeated_construction()
    test_str_repr()



"""
Package providing groups such as `NeuronGroup` or `PoissonGroup`.
"""

from .group import *
from .neurongroup import *
from .subgroup import *

__all__ = ["CodeRunner", "Group", "VariableOwner", "NeuronGroup"]



"""
A pseudo MSO neuron, with two dendrites and one axon (fake geometry).
"""
from brian2 import *

# Morphology
morpho = Soma(30*um)
morpho.axon = Cylinder(diameter=1*um, length=300*um, n=100)
morpho.L = Cylinder(diameter=1*um, length=100*um, n=50)
morpho.R = Cylinder(diameter=1*um, length=150*um, n=50)

# Passive channels
gL = 1e-4*siemens/cm**2
EL = -70*mV
eqs='''
Im = gL * (EL - v) : amp/meter**2
I : amp (point current)
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs,
                       Cm=1*uF/cm**2, Ri=100*ohm*cm, method='exponential_euler')
neuron.v = EL
neuron.I = 0*amp

# Monitors
mon_soma = StateMonitor(neuron, 'v', record=[0])
mon_L = StateMonitor(neuron.L, 'v', record=True)
mon_R = StateMonitor(neuron, 'v', record=morpho.R[75*um])

run(1*ms)
neuron.I[morpho.L[50*um]] = 0.2*nA  # injecting in the left dendrite
run(5*ms)
neuron.I = 0*amp
run(50*ms, report='text')

subplot(211)
plot(mon_L.t/ms, mon_soma[0].v/mV, 'k')
plot(mon_L.t/ms, mon_L[morpho.L[50*um]].v/mV, 'r')
plot(mon_L.t/ms, mon_R[morpho.R[75*um]].v/mV, 'b')
ylabel('v (mV)')
subplot(212)
for x in linspace(0*um, 100*um, 10, endpoint=False):
    plot(mon_L.t/ms, mon_L[morpho.L[x]].v/mV)
xlabel('Time (ms)')
ylabel('v (mV)')
show()



import ast
import numbers

import numpy as np
import sympy

from brian2.core.functions import DEFAULT_CONSTANTS, DEFAULT_FUNCTIONS

__all__ = [
    "NodeRenderer",
    "NumpyNodeRenderer",
    "CPPNodeRenderer",
    "SympyNodeRenderer",
]


class NodeRenderer:
    expression_ops = {
        # BinOp
        "Add": "+",
        "Sub": "-",
        "Mult": "*",
        "Div": "/",
        "FloorDiv": "//",
        "Pow": "**",
        "Mod": "%",
        # Compare
        "Lt": "<",
        "LtE": "<=",
        "Gt": ">",
        "GtE": ">=",
        "Eq": "==",
        "NotEq": "!=",
        # Unary ops
        "Not": "not",
        "UAdd": "+",
        "USub": "-",
        # Bool ops
        "And": "and",
        "Or": "or",
        # Augmented assign
        "AugAdd": "+=",
        "AugSub": "-=",
        "AugMult": "*=",
        "AugDiv": "/=",
        "AugPow": "**=",
        "AugMod": "%=",
    }

    def __init__(self, auto_vectorise=None):
        if auto_vectorise is None:
            auto_vectorise = set()
        self.auto_vectorise = auto_vectorise

    def render_expr(self, expr, strip=True):
        if strip:
            expr = expr.strip()
        node = ast.parse(expr, mode="eval")
        return self.render_node(node.body)

    def render_node(self, node):
        nodename = node.__class__.__name__
        methname = f"render_{nodename}"
        try:
            return getattr(self, methname)(node)
        except AttributeError:
            if nodename == "Subscript":
                raise SyntaxError(
                    "Brian equations/expressions do not support indexing with '[...]'."
                )
            elif nodename == "Attribute":
                raise SyntaxError(
                    "Brian equations/expressions do not support accessing attributes"
                    " with the '.' syntax."
                )
            elif nodename == "Tuple":
                raise SyntaxError("Brian equations/expressions do not support tuples.")
            else:
                raise SyntaxError(
                    f"Brian equations/expressions do not support the '{nodename}'"
                    " syntax."
                )

    def render_func(self, node):
        return self.render_Name(node)

    def render_Name(self, node):
        return node.id

    def render_Constant(self, node):
        if isinstance(node.value, np.number):
            # repr prints the dtype in numpy 2.0
            return repr(node.value.item())
        return repr(node.value)

    def render_Call(self, node):
        if len(node.keywords):
            raise ValueError("Keyword arguments not supported.")
        elif getattr(node, "starargs", None) is not None:
            raise ValueError("Variable number of arguments not supported")
        elif getattr(node, "kwargs", None) is not None:
            raise ValueError("Keyword arguments not supported")
        else:
            if node.func.id in self.auto_vectorise:
                vectorisation_idx = ast.Name("_vectorisation_idx")
                args = node.args + [vectorisation_idx]
            else:
                args = node.args
            return f"{self.render_func(node.func)}({', '.join(self.render_node(arg) for arg in args)})"

    def render_element_parentheses(self, node):
        """
        Render an element with parentheses around it or leave them away for
        numbers, names and function calls.
        """
        if node.__class__.__name__ == "Name":
            return self.render_node(node)
        elif node.__class__.__name__ in ["Num", "Constant"] and node.value >= 0:
            return self.render_node(node)
        elif node.__class__.__name__ == "Call":
            return self.render_node(node)
        else:
            return f"({self.render_node(node)})"

    def render_BinOp_parentheses(self, left, right, op):
        # Use a simplified checking whether it is possible to omit parentheses:
        # only omit parentheses for numbers, variable names or function calls.
        # This means we still put needless parentheses because we ignore
        # precedence rules, e.g. we write "3 + (4 * 5)" but at least we do
        # not do "(3) + ((4) + (5))"
        op_class = op.__class__.__name__
        # Give a more useful error message when using bit-wise operators
        if op_class in ["BitXor", "BitAnd", "BitOr"]:
            correction = {
                "BitXor": ("^", "**"),
                "BitAnd": ("&", "and"),
                "BitOr": ("|", "or"),
            }.get(op_class)
            raise SyntaxError(
                f'The operator "{correction[0]}" is not supported, use'
                f' "{correction[1]}" instead.'
            )
        return (
            f"{self.render_element_parentheses(left)} "
            f"{self.expression_ops[op_class]} "
            f"{self.render_element_parentheses(right)}"
        )

    def render_BinOp(self, node):
        return self.render_BinOp_parentheses(node.left, node.right, node.op)

    def render_BoolOp(self, node):
        op = self.expression_ops[node.op.__class__.__name__]
        return (f" {op} ").join(
            f"{self.render_element_parentheses(v)}" for v in node.values
        )

    def render_Compare(self, node):
        if len(node.comparators) > 1:
            raise SyntaxError("Can only handle single comparisons like a<b not a<b<c")
        return self.render_BinOp_parentheses(
            node.left, node.comparators[0], node.ops[0]
        )

    def render_UnaryOp(self, node):
        return f"{self.expression_ops[node.op.__class__.__name__]} {self.render_element_parentheses(node.operand)}"

    def render_Assign(self, node):
        if len(node.targets) > 1:
            raise SyntaxError("Only support syntax like a=b not a=b=c")
        return f"{self.render_node(node.targets[0])} = {self.render_node(node.value)}"

    def render_AugAssign(self, node):
        target = node.target.id
        rhs = self.render_node(node.value)
        op = self.expression_ops[f"Aug{node.op.__class__.__name__}"]
        return f"{target} {op} {rhs}"


class NumpyNodeRenderer(NodeRenderer):
    expression_ops = NodeRenderer.expression_ops.copy()
    expression_ops.update(
        {
            # Unary ops
            # We'll handle "not" explicitly below
            # Bool ops
            "And": "&",
            "Or": "|",
        }
    )

    def render_UnaryOp(self, node):
        if node.op.__class__.__name__ == "Not":
            return f"logical_not({self.render_node(node.operand)})"
        else:
            return NodeRenderer.render_UnaryOp(self, node)


class SympyNodeRenderer(NodeRenderer):
    expression_ops = {
        "Add": sympy.Add,
        "Mult": sympy.Mul,
        "Pow": sympy.Pow,
        "Mod": sympy.Mod,
        # Compare
        "Lt": sympy.StrictLessThan,
        "LtE": sympy.LessThan,
        "Gt": sympy.StrictGreaterThan,
        "GtE": sympy.GreaterThan,
        "Eq": sympy.Eq,
        "NotEq": sympy.Ne,
        # Unary ops are handled manually
        # Bool ops
        "And": sympy.And,
        "Or": sympy.Or,
    }

    def render_func(self, node):
        if node.id in DEFAULT_FUNCTIONS:
            f = DEFAULT_FUNCTIONS[node.id]
            if f.sympy_func is not None and isinstance(
                f.sympy_func, sympy.FunctionClass
            ):
                return f.sympy_func
        # special workaround for the "int" function
        if node.id == "int":
            return sympy.Function("int_")
        else:
            return sympy.Function(node.id)

    def render_Call(self, node):
        if len(node.keywords):
            raise ValueError("Keyword arguments not supported.")
        elif getattr(node, "starargs", None) is not None:
            raise ValueError("Variable number of arguments not supported")
        elif getattr(node, "kwargs", None) is not None:
            raise ValueError("Keyword arguments not supported")
        elif len(node.args) == 0:
            return self.render_func(node.func)(sympy.Symbol("_placeholder_arg"))
        else:
            return self.render_func(node.func)(
                *(self.render_node(arg) for arg in node.args)
            )

    def render_Compare(self, node):
        if len(node.comparators) > 1:
            raise SyntaxError("Can only handle single comparisons like a<b not a<b<c")
        op = node.ops[0]
        return self.expression_ops[op.__class__.__name__](
            self.render_node(node.left), self.render_node(node.comparators[0])
        )

    def render_Name(self, node):
        if node.id in DEFAULT_CONSTANTS:
            c = DEFAULT_CONSTANTS[node.id]
            return c.sympy_obj
        elif node.id in ["t", "dt"]:
            return sympy.Symbol(node.id, real=True, positive=True)
        else:
            return sympy.Symbol(node.id, real=True)

    def render_Constant(self, node):
        if node.value is True or node.value is False:
            return node.value
        elif isinstance(node.value, numbers.Integral):
            return sympy.Integer(node.value)
        elif isinstance(node.value, numbers.Number):
            return sympy.Float(node.value)
        else:
            return str(node.value)

    def render_BinOp(self, node):
        op_name = node.op.__class__.__name__
        # Sympy implements division and subtraction as multiplication/addition
        if op_name == "Div":
            op = self.expression_ops["Mult"]
            return op(self.render_node(node.left), 1 / self.render_node(node.right))
        elif op_name == "FloorDiv":
            op = self.expression_ops["Mult"]
            left = self.render_node(node.left)
            right = self.render_node(node.right)
            return sympy.floor(op(left, 1 / right))
        elif op_name == "Sub":
            op = self.expression_ops["Add"]
            return op(self.render_node(node.left), -self.render_node(node.right))
        else:
            op = self.expression_ops[op_name]
            return op(self.render_node(node.left), self.render_node(node.right))

    def render_BoolOp(self, node):
        op = self.expression_ops[node.op.__class__.__name__]
        return op(*(self.render_node(value) for value in node.values))

    def render_UnaryOp(self, node):
        op_name = node.op.__class__.__name__
        if op_name == "UAdd":
            # Nothing to do
            return self.render_node(node.operand)
        elif op_name == "USub":
            return -self.render_node(node.operand)
        elif op_name == "Not":
            return sympy.Not(self.render_node(node.operand))
        else:
            raise ValueError(f"Unknown unary operator: {op_name}")


class CPPNodeRenderer(NodeRenderer):
    expression_ops = NodeRenderer.expression_ops.copy()
    expression_ops.update(
        {
            # Unary ops
            "Not": "!",
            # Bool ops
            "And": "&&",
            "Or": "||",
            # C does not have a floor division operator (but see render_BinOp)
            "FloorDiv": "/",
        }
    )

    def render_BinOp(self, node):
        if node.op.__class__.__name__ == "Pow":
            return (
                f"_brian_pow({self.render_node(node.left)},"
                f" {self.render_node(node.right)})"
            )
        elif node.op.__class__.__name__ == "Mod":
            return (
                f"_brian_mod({self.render_node(node.left)},"
                f" {self.render_node(node.right)})"
            )
        elif node.op.__class__.__name__ == "Div":
            # C uses integer division, this is a quick and dirty way to assure
            # it uses floating point division for integers
            return f"1.0f*{self.render_element_parentheses(node.left)}/{self.render_element_parentheses(node.right)}"
        elif node.op.__class__.__name__ == "FloorDiv":
            return (
                f"_brian_floordiv({self.render_node(node.left)},"
                f" {self.render_node(node.right)})"
            )
        else:
            return NodeRenderer.render_BinOp(self, node)

    def render_Constant(self, node):
        if node.value is True:
            return "true"
        elif node.value is False:
            return "false"
        else:
            return super().render_Constant(node)

    def render_Name(self, node):
        if node.id == "inf":
            return "INFINITY"
        else:
            return node.id

    def render_Assign(self, node):
        return f"{NodeRenderer.render_Assign(self, node)};"



"""
This module translates a series of statements into a language-specific
syntactically correct code block that can be inserted into a template.

It infers whether or not a variable can be declared as
constant, etc. It should handle common subexpressions, and so forth.

The input information needed:

* The sequence of statements (a multiline string) in standard mathematical form
* The list of known variables, common subexpressions and functions, and for each
  variable whether or not it is a value or an array, and if an array what the
  dtype is.
* The dtype to use for newly created variables
* The language to translate to
"""

import re
from collections.abc import Mapping

import numpy as np
import sympy

from brian2.core.functions import Function
from brian2.core.preferences import prefs
from brian2.core.variables import AuxiliaryVariable, Subexpression, Variable
from brian2.parsing.bast import brian_ast
from brian2.parsing.statements import parse_statement
from brian2.parsing.sympytools import str_to_sympy, sympy_to_str
from brian2.utils.caching import cached
from brian2.utils.stringtools import deindent, get_identifiers, strip_empty_lines
from brian2.utils.topsort import topsort

from .optimisation import optimise_statements
from .statements import Statement

__all__ = ["analyse_identifiers", "get_identifiers_recursively"]


class LineInfo:
    """
    A helper class, just used to store attributes.
    """

    def __init__(self, **kwds):
        for k, v in kwds.items():
            setattr(self, k, v)

    # TODO: This information should go somewhere else, I guess


STANDARD_IDENTIFIERS = {"and", "or", "not", "True", "False"}


def analyse_identifiers(code, variables, recursive=False):
    """
    Analyses a code string (sequence of statements) to find all identifiers by type.

    In a given code block, some variable names (identifiers) must be given as inputs to the code
    block, and some are created by the code block. For example, the line::

        a = b+c

    This could mean to create a new variable a from b and c, or it could mean modify the existing
    value of a from b or c, depending on whether a was previously known.

    Parameters
    ----------
    code : str
        The code string, a sequence of statements one per line.
    variables : dict of `Variable`, set of names
        Specifiers for the model variables or a set of known names
    recursive : bool, optional
        Whether to recurse down into subexpressions (defaults to ``False``).

    Returns
    -------
    newly_defined : set
        A set of variables that are created by the code block.
    used_known : set
        A set of variables that are used and already known, a subset of the
        ``known`` parameter.
    unknown : set
        A set of variables which are used by the code block but not defined by
        it and not previously known. Should correspond to variables in the
        external namespace.
    """
    if isinstance(variables, Mapping):
        known = {
            k for k, v in variables.items() if not isinstance(k, AuxiliaryVariable)
        }
    else:
        known = set(variables)
        variables = {k: Variable(name=k, dtype=np.float64) for k in known}

    known |= STANDARD_IDENTIFIERS
    scalar_stmts, vector_stmts = make_statements(
        code, variables, np.float64, optimise=False
    )
    stmts = scalar_stmts + vector_stmts
    defined = {stmt.var for stmt in stmts if stmt.op == ":="}
    if len(stmts) == 0:
        allids = set()
    elif recursive:
        if not isinstance(variables, Mapping):
            raise TypeError("Have to specify a variables dictionary.")
        allids = get_identifiers_recursively(
            [stmt.expr for stmt in stmts], variables
        ) | {stmt.var for stmt in stmts}
    else:
        allids = set.union(*[get_identifiers(stmt.expr) for stmt in stmts]) | {
            stmt.var for stmt in stmts
        }
    dependent = allids.difference(defined, known)
    used_known = allids.intersection(known) - STANDARD_IDENTIFIERS
    return defined, used_known, dependent


def get_identifiers_recursively(expressions, variables, include_numbers=False):
    """
    Gets all the identifiers in a list of expressions, recursing down into
    subexpressions.

    Parameters
    ----------
    expressions : list of str
        List of expressions to check.
    variables : dict-like
        Dictionary of `Variable` objects
    include_numbers : bool, optional
        Whether to include number literals in the output. Defaults to ``False``.
    """
    if len(expressions):
        identifiers = set.union(
            *[
                get_identifiers(expr, include_numbers=include_numbers)
                for expr in expressions
            ]
        )
    else:
        identifiers = set()
    for name in set(identifiers):
        if name in variables and isinstance(variables[name], Subexpression):
            s_identifiers = get_identifiers_recursively(
                [variables[name].expr], variables, include_numbers=include_numbers
            )
            identifiers |= s_identifiers
    return identifiers


def is_scalar_expression(expr, variables):
    """
    Whether the given expression is scalar.

    Parameters
    ----------
    expr : str
        The expression to check
    variables : dict-like
        `Variable` and `Function` object for all the identifiers used in `expr`

    Returns
    -------
    scalar : bool
        Whether `expr` is a scalar expression
    """
    # determine whether this is a scalar variable
    identifiers = get_identifiers_recursively([expr], variables)
    # In the following we assume that all unknown identifiers are
    # scalar constants -- this should cover numerical literals and
    # e.g. "True" or "inf".
    return all(
        name not in variables
        or getattr(variables[name], "scalar", False)
        or (isinstance(variables[name], Function) and variables[name].stateless)
        for name in identifiers
    )


@cached
def make_statements(code, variables, dtype, optimise=True, blockname=""):
    """
    make_statements(code, variables, dtype, optimise=True, blockname='')

    Turn a series of abstract code statements into Statement objects, inferring
    whether each line is a set/declare operation, whether the variables are
    constant or not, and handling the cacheing of subexpressions.

    Parameters
    ----------
    code : str
        A (multi-line) string of statements.
    variables : dict-like
        A dictionary of with `Variable` and `Function` objects for every
        identifier used in the `code`.
    dtype : `dtype`
        The data type to use for temporary variables
    optimise : bool, optional
        Whether to optimise expressions, including
        pulling out loop invariant expressions and putting them in new
        scalar constants. Defaults to ``False``, since this function is also
        used just to in contexts where we are not interested by this kind of
        optimisation. For the main code generation stage, its value is set by
        the `codegen.loop_invariant_optimisations` preference.
    blockname : str, optional
        A name for the block (used to name intermediate variables to avoid
        name clashes when multiple blocks are used together)
    Returns
    -------
    scalar_statements, vector_statements : (list of `Statement`, list of `Statement`)
        Lists with statements that are to be executed once and statements that
        are to be executed once for every neuron/synapse/... (or in a vectorised
        way)

    Notes
    -----
    If ``optimise`` is ``True``, then the
    ``scalar_statements`` may include newly introduced scalar constants that
    have been identified as loop-invariant and have therefore been pulled out
    of the vector statements. The resulting statements will also use augmented
    assignments where possible, i.e. a statement such as ``w = w + 1`` will be
    replaced by ``w += 1``. Also, statements involving booleans will have
    additional information added to them (see `Statement` for details)
    describing how the statement can be reformulated as a sequence of if/then
    statements. Calls `~brian2.codegen.optimisation.optimise_statements`.
    """
    code = strip_empty_lines(deindent(code))
    lines = re.split(r"[;\n]", code)
    lines = [LineInfo(code=line) for line in lines if len(line)]
    # Do a copy so we can add stuff without altering the original dict
    variables = dict(variables)
    # we will do inference to work out which lines are := and which are =
    defined = {k for k, v in variables.items() if not isinstance(v, AuxiliaryVariable)}
    for line in lines:
        statement = None
        # parse statement into "var op expr"
        var, op, expr, comment = parse_statement(line.code)
        if var in variables and isinstance(variables[var], Subexpression):
            raise SyntaxError(
                f"Illegal line '{line.code}' in abstract code. Cannot write to"
                f" subexpression '{var}'."
            )
        if op == "=":
            if var not in defined:
                op = ":="
                defined.add(var)
                if var not in variables:
                    annotated_ast = brian_ast(expr, variables)
                    is_scalar = annotated_ast.scalar
                    if annotated_ast.dtype == "boolean":
                        use_dtype = bool
                    elif annotated_ast.dtype == "integer":
                        use_dtype = int
                    else:
                        use_dtype = dtype
                    new_var = AuxiliaryVariable(var, dtype=use_dtype, scalar=is_scalar)
                    variables[var] = new_var
            elif not variables[var].is_boolean:
                sympy_expr = str_to_sympy(expr, variables)
                if variables[var].is_integer:
                    sympy_var = sympy.Symbol(var, integer=True)
                else:
                    sympy_var = sympy.Symbol(var, real=True)
                try:
                    collected = sympy.collect(
                        sympy_expr, sympy_var, exact=True, evaluate=False
                    )
                except AttributeError:
                    # If something goes wrong during collection, e.g. collect
                    # does not work for logical expressions
                    collected = {1: sympy_expr}

                if (
                    len(collected) == 2
                    and set(collected.keys()) == {1, sympy_var}
                    and collected[sympy_var] == 1
                ):
                    # We can replace this statement by a += assignment
                    statement = Statement(
                        var,
                        "+=",
                        sympy_to_str(collected[1]),
                        comment,
                        dtype=variables[var].dtype,
                        scalar=variables[var].scalar,
                    )
                elif len(collected) == 1 and sympy_var in collected:
                    # We can replace this statement by a *= assignment
                    statement = Statement(
                        var,
                        "*=",
                        sympy_to_str(collected[sympy_var]),
                        comment,
                        dtype=variables[var].dtype,
                        scalar=variables[var].scalar,
                    )
        if statement is None:
            statement = Statement(
                var,
                op,
                expr,
                comment,
                dtype=variables[var].dtype,
                scalar=variables[var].scalar,
            )

        line.statement = statement
        # for each line will give the variable being written to
        line.write = var
        # each line will give a set of variables which are read
        line.read = get_identifiers_recursively([expr], variables)

    # All writes to scalar variables must happen before writes to vector
    # variables
    scalar_write_done = False
    for line in lines:
        stmt = line.statement
        if stmt.op != ":=" and variables[stmt.var].scalar and scalar_write_done:
            raise SyntaxError(
                "All writes to scalar variables in a code block "
                "have to be made before writes to vector "
                f"variables. Illegal write to '{line.write}'."
            )
        elif not variables[stmt.var].scalar:
            scalar_write_done = True

    # backwards compute whether or not variables will be read again
    # note that will_read for a line gives the set of variables it will read
    # on the current line or subsequent ones. will_write gives the set of
    # variables that will be written after the current line
    will_read = set()
    will_write = set()
    for line in lines[::-1]:
        will_read = will_read.union(line.read)
        line.will_read = will_read.copy()
        line.will_write = will_write.copy()
        will_write.add(line.write)

    subexpressions = {
        name: val for name, val in variables.items() if isinstance(val, Subexpression)
    }
    # Check that no scalar subexpression refers to a vectorised function
    # (e.g. rand()) -- otherwise it would be differently interpreted depending
    # on whether it is used in a scalar or a vector context (i.e., even though
    # the subexpression is supposed to be scalar, it would be vectorised when
    # used as part of non-scalar expressions)
    for name, subexpr in subexpressions.items():
        if subexpr.scalar:
            identifiers = get_identifiers(subexpr.expr)
            for identifier in identifiers:
                if identifier in variables and getattr(
                    variables[identifier], "auto_vectorise", False
                ):
                    raise SyntaxError(
                        f"The scalar subexpression '{name}' refers to "
                        f"the implicitly vectorised function '{identifier}' "
                        "-- this is not allowed since it leads "
                        "to different interpretations of this "
                        "subexpression depending on whether it "
                        "is used in a scalar or vector "
                        "context."
                    )

    # sort subexpressions into an order so that subexpressions that don't depend
    # on other subexpressions are first
    subexpr_deps = {
        name: [dep for dep in subexpr.identifiers if dep in subexpressions]
        for name, subexpr in subexpressions.items()
    }
    sorted_subexpr_vars = topsort(subexpr_deps)

    statements = []

    # none are yet defined (or declared)
    subdefined = {name: None for name in subexpressions}
    for line in lines:
        # update/define all subexpressions needed by this statement
        for var in sorted_subexpr_vars:
            if var not in line.read:
                continue

            subexpression = subexpressions[var]
            # if already defined/declared
            if subdefined[var] == "constant":
                continue
            elif subdefined[var] == "variable":
                op = "="
                constant = False
            else:
                op = ":="
                # check if the referred variables ever change
                ids = subexpression.identifiers
                constant = all(v not in line.will_write for v in ids)
                subdefined[var] = "constant" if constant else "variable"

            statement = Statement(
                var,
                op,
                subexpression.expr,
                comment="",
                dtype=variables[var].dtype,
                constant=constant,
                subexpression=True,
                scalar=variables[var].scalar,
            )
            statements.append(statement)

        stmt = line.statement
        var, op, expr, comment = stmt.var, stmt.op, stmt.expr, stmt.comment

        # constant only if we are declaring a new variable and we will not
        # write to it again
        constant = op == ":=" and var not in line.will_write
        statement = Statement(
            var,
            op,
            expr,
            comment,
            dtype=variables[var].dtype,
            constant=constant,
            scalar=variables[var].scalar,
        )
        statements.append(statement)

    scalar_statements = [s for s in statements if s.scalar]
    vector_statements = [s for s in statements if not s.scalar]

    if optimise and prefs.codegen.loop_invariant_optimisations:
        scalar_statements, vector_statements = optimise_statements(
            scalar_statements, vector_statements, variables, blockname=blockname
        )

    return scalar_statements, vector_statements



"""
Tests the brian2.parsing package
"""

from collections import namedtuple

import numpy as np
import pytest

from brian2 import Function
from brian2.codegen.generators.cpp_generator import CPPCodeGenerator
from brian2.core.functions import DEFAULT_FUNCTIONS
from brian2.core.preferences import prefs
from brian2.core.variables import Constant
from brian2.groups.group import Group
from brian2.parsing.dependencies import abstract_code_dependencies
from brian2.parsing.expressions import (
    _get_value_from_expression,
    is_boolean_expression,
    parse_expression_dimensions,
)
from brian2.parsing.functions import (
    abstract_code_from_function,
    extract_abstract_code_functions,
    substitute_abstract_code_functions,
)
from brian2.parsing.rendering import CPPNodeRenderer, NodeRenderer, NumpyNodeRenderer
from brian2.parsing.sympytools import str_to_sympy, sympy_to_str
from brian2.tests.utils import assert_allclose
from brian2.units import (
    DimensionMismatchError,
    Unit,
    amp,
    get_unit,
    have_same_dimensions,
    volt,
)
from brian2.units.fundamentalunits import DIMENSIONLESS, Dimension
from brian2.utils.logger import std_silent
from brian2.utils.stringtools import deindent, get_identifiers


# a simple Group for testing
class SimpleGroup(Group):
    def __init__(self, variables, namespace=None):
        self.variables = variables
        self.namespace = namespace


TEST_EXPRESSIONS = """
    a+b+c*d-f+g-(b+d)-(a-c)
    a**b**2
    a**(b**2)
    (a**b)**2
    a*(b+c*(a+b)*(a-(c*d)))
    a/b/c-a/(b/c)
    10//n
    n//10
    n//m
    10/n
    10.0/n
    n/10
    n/10.0
    n/m
    a<b
    a<=b
    a>b
    a>=b
    a==b
    a!=b
    a+1
    1+a
    1+3
    a>0.5 and b>0.5
    a>0.5 and b>0.5 or c>0.5
    a>0.5 and b>0.5 or not c>0.5
    2%4
    -1%4
    2.3%5.6
    2.3%5
    -1.2%3.4
    17e-12
    42e17
    """


def parse_expressions(renderer, evaluator, numvalues=10):
    exprs = [
        ([m for m in get_identifiers(l) if len(m) == 1], [], l.strip())
        for l in TEST_EXPRESSIONS.split("\n")
        if l.strip()
    ]
    i, imod = 1, 33
    for varids, funcids, expr in exprs:
        pexpr = renderer.render_expr(expr)
        n = 0
        for _ in range(numvalues):
            # assign some random values
            ns = {}
            for v in varids:
                if v in ["n", "m"]:  # integer values
                    ns[v] = i
                else:
                    ns[v] = float(i) / imod
                i = i % imod + 1
            r1 = eval(expr.replace("&", " and ").replace("|", " or "), ns)
            n += 1
            r2 = evaluator(pexpr, ns)
            try:
                # Use all close because we can introduce small numerical
                # difference through sympy's rearrangements
                assert_allclose(r1, r2, atol=1e-8)
            except AssertionError as e:
                raise AssertionError(
                    f"In expression {str(expr)} translated to {str(pexpr)} {str(e)}"
                )


def numpy_evaluator(expr, userns):
    ns = {}
    # exec 'from numpy import logical_not' in ns
    ns["logical_not"] = np.logical_not
    ns.update(**userns)
    for k in userns:
        if not k.startswith("_"):
            ns[k] = np.array([userns[k]])
    try:
        x = eval(expr, ns)
    except Exception as e:
        raise ValueError(
            f"Could not evaluate numpy expression {expr} exception {str(e)}"
        )
    if isinstance(x, np.ndarray):
        return x[0]
    else:
        return x


@pytest.mark.codegen_independent
def test_parse_expressions_python():
    parse_expressions(NodeRenderer(), eval)


@pytest.mark.codegen_independent
def test_parse_expressions_numpy():
    parse_expressions(NumpyNodeRenderer(), numpy_evaluator)


@pytest.mark.codegen_independent
def test_parse_expressions_sympy():
    # sympy is about symbolic calculation, the string returned by the renderer
    # contains "Symbol('a')" etc. so we cannot simply evaluate it in a
    # namespace.
    # We therefore use a different approach: Convert the expression to a
    # sympy expression via str_to_sympy (uses the SympyNodeRenderer internally),
    # then convert it back to a string via sympy_to_str and evaluate it

    class SympyRenderer:
        def render_expr(self, expr):
            return str_to_sympy(expr)

    def evaluator(expr, ns):
        expr = sympy_to_str(expr)
        ns = dict(ns)
        # Add the floor function which is used to implement floor division
        ns["floor"] = DEFAULT_FUNCTIONS["floor"]
        return eval(expr, ns)

    parse_expressions(SympyRenderer(), evaluator)


@pytest.mark.codegen_independent
def test_abstract_code_dependencies():
    code = """
    a = b+c
    d = b+c
    a = func_a()
    a = func_b()
    a = x+d
    """
    known_vars = {"a", "b", "c"}
    known_funcs = {"func_a"}
    res = abstract_code_dependencies(code, known_vars, known_funcs)
    expected_res = dict(
        all=[
            "a",
            "b",
            "c",
            "d",
            "x",
            "func_a",
            "func_b",
        ],
        read=["b", "c", "d", "x"],
        write=["a", "d"],
        funcs=["func_a", "func_b"],
        known_all=["a", "b", "c", "func_a"],
        known_read=["b", "c"],
        known_write=["a"],
        known_funcs=["func_a"],
        unknown_read=["d", "x"],
        unknown_write=["d"],
        unknown_funcs=["func_b"],
        undefined_read=["x"],
        newly_defined=["d"],
    )
    for k, v in expected_res.items():
        if not getattr(res, k) == set(v):
            raise AssertionError(
                f"For '{k}' result is {getattr(res, k)} expected {set(v)}"
            )


@pytest.mark.codegen_independent
def test_is_boolean_expression():
    # dummy "Variable" class
    Var = namedtuple("Var", ["is_boolean"])

    # dummy function object
    class Func:
        def __init__(self, returns_bool=False):
            self._returns_bool = returns_bool

    # variables / functions
    a = Constant("a", value=True)
    b = Constant("b", value=False)
    c = Constant("c", value=5)
    f = Func(returns_bool=True)
    g = Func(returns_bool=False)
    s1 = Var(is_boolean=True)
    s2 = Var(is_boolean=False)

    variables = {"a": a, "b": b, "c": c, "f": f, "g": g, "s1": s1, "s2": s2}

    EVF = [
        (True, "a or b"),
        (False, "c"),
        (False, "s2"),
        (False, "g(s1)"),
        (True, "s2 > c"),
        (True, "c > 5"),
        (True, "True"),
        (True, "a<b"),
        (True, "not (a>=b)"),
        (False, "a+b"),
        (True, "f(c)"),
        (False, "g(c)"),
        (
            True,
            "f(c) or a<b and s1",
        ),
    ]
    for expect, expr in EVF:
        ret_val = is_boolean_expression(expr, variables)
        if expect != ret_val:
            raise AssertionError(
                f"is_boolean_expression({expr!r}) returned 'ret_val', "
                "but was supposed to return 'expect'."
            )
    with pytest.raises(SyntaxError):
        is_boolean_expression("a<b and c", variables)
    with pytest.raises(SyntaxError):
        is_boolean_expression("a or foo", variables)
    with pytest.raises(SyntaxError):
        is_boolean_expression("ot a", variables)  # typo
    with pytest.raises(SyntaxError):
        is_boolean_expression("g(c) and f(a)", variables)


@pytest.mark.codegen_independent
@pytest.mark.parametrize(
    "expect,expr",
    [
        (volt * amp, "a+b*c"),
        (DimensionMismatchError, "a+b"),
        (DimensionMismatchError, "a<b"),
        (1, "a<b*c"),
        (1, "a or b"),
        (1, "not (a >= b*c)"),
        (DimensionMismatchError, "a or b<c"),
        (1, "a/(b*c)<1"),
        (1, "a/(a-a)"),
        (1, "a<mV*mA"),
        (volt**2, "b**2"),
        (volt * amp, "a%(b*c)"),
        (volt, "-b"),
        (1, "(a/a)**(a/a)"),
        # Expressions involving functions
        (volt, "rand()*b"),
        (volt**0.5, "sqrt(b)"),
        (volt, "ceil(b)"),
        (volt, "sqrt(randn()*b**2)"),
        (1, "sin(b/b)"),
        (DimensionMismatchError, "sin(b)"),
        (DimensionMismatchError, "sqrt(b) + b"),
        (SyntaxError, "sqrt(b, b)"),
        (SyntaxError, "sqrt()"),
        (SyntaxError, "int(1, 2)"),
    ],
)
def test_parse_expression_unit(expect, expr):
    Var = namedtuple("Var", ["dim", "dtype"])
    variables = {
        "a": Var(dim=(volt * amp).dim, dtype=np.float64),
        "b": Var(dim=volt.dim, dtype=np.float64),
        "c": Var(dim=amp.dim, dtype=np.float64),
    }
    group = SimpleGroup(namespace={}, variables=variables)
    all_variables = {}
    for name in get_identifiers(expr):
        if name in variables:
            all_variables[name] = variables[name]
        else:
            all_variables[name] = group._resolve(name, {})

    if isinstance(expect, type) and issubclass(expect, Exception):
        with pytest.raises(expect):
            parse_expression_dimensions(expr, all_variables)
    else:
        u = parse_expression_dimensions(expr, all_variables)
        assert have_same_dimensions(u, expect)


@pytest.mark.codegen_independent
@pytest.mark.parametrize("expr", ["a**b", "a << b", "int(True"])  # typo
def test_parse_expression_unit_wrong_expressions(expr):
    Var = namedtuple("Var", ["dim", "dtype"])
    variables = {
        "a": Var(dim=(volt * amp).dim, dtype=np.float64),
        "b": Var(dim=volt.dim, dtype=np.float64),
        "c": Var(dim=amp.dim, dtype=np.float64),
    }
    all_variables = {}
    group = SimpleGroup(namespace={}, variables=variables)
    for name in get_identifiers(expr):
        if name in variables:
            all_variables[name] = variables[name]
        else:
            all_variables[name] = group._resolve(name, {})
    with pytest.raises(SyntaxError):
        parse_expression_dimensions(expr, all_variables)


@pytest.mark.codegen_independent
@pytest.mark.parametrize(
    "expr,correct",
    [
        ("sin(5)", True),
        ("3 + sin(5)", True),
        ("d + sin(5)", True),
        ("3 + sin(d)", True),
        ("sin(5*mV)", False),
        ("sin(a)", False),
        ("3*mV + sin(5)", False),
        ("b + sin(5)", False),
        ("sqrt(b**2) + sqrt(7*mV**2)", True),
        ("sqrt(d) + sqrt(7*mV**2)", False),
        ("b + clip(3*mV, 0*mV, 5*mV)", True),
        ("b + clip(3*mV, 0, 5)", False),
        ("b + foo(7, 3*mV, 9)", True),
        ("a + foo(7*nA, 3*mV, 9*nA)", True),
        ("b + foo(7, 3*mV, 9*mV)", False),
        ("a + foo(7*nA, 3*mV, 9)", False),
    ],
)
def test_parse_expression_unit_functions(expr, correct):
    Var = namedtuple("Var", ["dim", "dtype"])

    def foo(x, y, z):
        return (x + z) * y

    variables = {
        "a": Var(dim=(volt * amp).dim, dtype=np.float64),
        "b": Var(dim=volt.dim, dtype=np.float64),
        "c": Var(dim=amp.dim, dtype=np.float64),
        "d": Var(dim=DIMENSIONLESS, dtype=np.float64),
        "foo": Function(
            pyfunc=foo,
            arg_units=[None, volt, "x"],
            arg_names=["x", "y", "z"],
            return_unit=lambda x, y, z: x * y,
        ),
    }
    all_variables = {}
    group = SimpleGroup(namespace={}, variables=variables)
    for name in get_identifiers(expr):
        if name in variables:
            all_variables[name] = variables[name]
        else:
            all_variables[name] = group._resolve(name, {})
    if correct:
        assert isinstance(parse_expression_dimensions(expr, all_variables), Dimension)
    else:
        with pytest.raises(DimensionMismatchError):
            parse_expression_dimensions(expr, all_variables)


@pytest.mark.codegen_independent
def test_value_from_expression():
    # This function is used to get the value of an exponent, necessary for unit checking

    constants = {"c": 3}

    # dummy class
    class C:
        pass

    variables = {"s_constant_scalar": C(), "s_non_constant": C(), "s_non_scalar": C()}
    variables["s_constant_scalar"].scalar = True
    variables["s_constant_scalar"].constant = True
    variables["s_constant_scalar"].get_value = lambda: 2.0
    variables["s_non_scalar"].constant = True
    variables["s_non_constant"].scalar = True
    variables["c"] = Constant("c", value=3)

    expressions = [
        "1",
        "-0.5",
        "c",
        "2**c",
        "(c + 3) * 5",
        "c + s_constant_scalar",
        "True",
        "False",
    ]

    for expr in expressions:
        eval_expr = expr.replace("s_constant_scalar", "s_constant_scalar.get_value()")
        assert float(
            eval(eval_expr, variables, constants)
        ) == _get_value_from_expression(expr, variables)

    wrong_expressions = ["s_non_constant", "s_non_scalar", "c or True"]
    for expr in wrong_expressions:
        with pytest.raises(SyntaxError):
            _get_value_from_expression(expr, variables)


@pytest.mark.codegen_independent
def test_abstract_code_from_function():
    # test basic functioning
    def f(x):
        y = x + 1
        return y * y

    ac = abstract_code_from_function(f)
    assert ac.name == "f"
    assert ac.args == ["x"]
    assert ac.code.strip() == "y = x + 1"
    assert ac.return_expr == "y * y"
    # Check that unsupported features raise an error

    def f(x):
        return x[:]

    with pytest.raises(SyntaxError):
        abstract_code_from_function(f)

    def f(x, **kwarg):
        return x

    with pytest.raises(SyntaxError):
        abstract_code_from_function(f)

    def f(x, *args):
        return x

    with pytest.raises(SyntaxError):
        abstract_code_from_function(f)


@pytest.mark.codegen_independent
def test_extract_abstract_code_functions():
    code = """
    def f(x):
        return x*x
        
    def g(V):
        V += 1
        
    irrelevant_code_here()
    """
    funcs = extract_abstract_code_functions(code)
    assert funcs["f"].return_expr == "x * x"
    assert funcs["g"].args == ["V"]


@pytest.mark.codegen_independent
def test_substitute_abstract_code_functions():
    def f(x):
        y = x * x
        return y

    def g(x):
        return f(x) + 1

    code = """
    z = f(x)
    z = f(x)+f(y)
    w = f(z)
    h = f(f(w))
    p = g(g(x))
    """
    funcs = [
        abstract_code_from_function(f),
        abstract_code_from_function(g),
    ]
    subcode = substitute_abstract_code_functions(code, funcs)
    for x, y in [(0, 1), (1, 0), (0.124323, 0.4549483)]:
        ns1 = {"x": x, "y": y, "f": f, "g": g}
        ns2 = {"x": x, "y": y}
        exec(deindent(code), ns1)
        exec(subcode, ns2)
        for k in ["z", "w", "h", "p"]:
            assert ns1[k] == ns2[k]


@pytest.mark.codegen_independent
def test_sympytools():
    # sympy_to_str(str_to_sympy(x)) should equal x

    # Note that the test below is quite fragile since sympy might rearrange the
    # order of symbols
    expressions = [
        "randn()",  # argumentless function
        "x + sin(pi*freq*t)",  # expression with a constant
        "c * userfun(t + x)",  # non-sympy function
        "abs(x) + ceil(y)",  # functions with a different name in sympy
        "inf",  # constant with a different name in sympy
        "not(b)",  # boolean expression
    ]

    for expr in expressions:
        expr2 = sympy_to_str(str_to_sympy(expr))
        assert expr.replace(" ", "") == expr2.replace(" ", ""), f"{expr} != {expr2}"


@pytest.mark.codegen_independent
def test_error_messages():
    nr = NodeRenderer()
    expr_expected = [
        ("3^2", "^", "**"),
        ("int(not_refractory | (v > 30))", "|", "or"),
        ("int((v > 30) & (w < 20))", "&", "and"),
        ("x +* 3", "", ""),
        ("v[index]", "indexing", ""),
        ("v.value", "attribute", ""),
        ("(v, w)", "tuple", ""),
    ]
    for expr, expected_1, expected_2 in expr_expected:
        try:
            nr.render_expr(expr)
            raise AssertionError(f"Excepted {expr} to raise a SyntaxError.")
        except SyntaxError as exc:
            message = str(exc)
            assert expected_1 in message
            assert expected_2 in message


@pytest.mark.codegen_independent
def test_sympy_infinity():
    # See github issue #1061
    assert sympy_to_str(str_to_sympy("inf")) == "inf"
    assert sympy_to_str(str_to_sympy("-inf")) == "-inf"


if __name__ == "__main__":
    from _pytest.outcomes import Skipped

    test_parse_expressions_python()
    test_parse_expressions_numpy()
    try:
        test_parse_expressions_cpp()
    except Skipped:
        pass
    test_parse_expressions_sympy()
    test_abstract_code_dependencies()
    test_is_boolean_expression()
    test_parse_expression_unit()
    test_value_from_expression()
    test_abstract_code_from_function()
    test_extract_abstract_code_functions()
    test_substitute_abstract_code_functions()
    test_sympytools()
    test_error_messages()
    test_sympy_infinity()



"""
Module containing the `Device` base class as well as the `RuntimeDevice`
implementation and some helper functions to access/set devices.
"""

import numbers
from weakref import WeakKeyDictionary

import numpy as np

from brian2.codegen.targets import codegen_targets
from brian2.core.names import find_name
from brian2.core.preferences import prefs
from brian2.core.variables import ArrayVariable, DynamicArrayVariable
from brian2.memory.dynamicarray import DynamicArray, DynamicArray1D
from brian2.units import ms
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import code_representation, indent

__all__ = [
    "Device",
    "RuntimeDevice",
    "get_device",
    "set_device",
    "all_devices",
    "reinit_devices",
    "reinit_and_delete",
    "reset_device",
    "device",
    "seed",
]

logger = get_logger(__name__)

all_devices = {}


prefs.register_preferences("devices", "Device preferences")


#: caches the automatically determined code generation target
_auto_target = None


def auto_target():
    """
    Automatically chose a code generation target (invoked when the
    `codegen.target` preference is set to `'auto'`. Caches its result so it
    only does the check once. Prefers cython > numpy.

    Returns
    -------
    target : class derived from `CodeObject`
        The target to use
    """
    global _auto_target
    if _auto_target is None:
        target_dict = {
            target.class_name: target for target in codegen_targets if target.class_name
        }
        using_fallback = False
        if "cython" in target_dict and target_dict["cython"].is_available():
            _auto_target = target_dict["cython"]
        else:
            _auto_target = target_dict["numpy"]
            using_fallback = True

        if using_fallback:
            logger.info(
                "Cannot use compiled code, falling back to the numpy "
                "code generation target. Note that this will likely "
                "be slower than using compiled code. Set the code "
                "generation to numpy manually to avoid this message:\n"
                'prefs.codegen.target = "numpy"',
                "codegen_fallback",
                once=True,
            )
        else:
            logger.debug(
                "Chosing %r as the code generation target." % _auto_target.class_name
            )

    return _auto_target


class Device:
    """
    Base Device object.
    """

    def __init__(self):
        #: The network schedule that this device supports. If the device only
        #: supports a specific, fixed schedule, it has to set this attribute to
        #: the respective schedule (see `Network.schedule` for details). If it
        #: supports arbitrary schedules, it should be set to ``None`` (the
        #: default).
        self.network_schedule = None

        self.defaultclock = None

        self._maximum_run_time = None

        self._state_tuple = (self.__module__, self.__class__.__name__)

    def _set_maximum_run_time(self, maximum_run_time):
        """
        Sets a maximum time for a run before it will break. Used primarily for testing purposes. Not guaranteed to be
        respected by a device.
        """
        self._maximum_run_time = maximum_run_time

    def get_array_name(self, var, access_data=True):
        """
        Return a globally unique name for `var`.

        Parameters
        ----------
        access_data : bool, optional
            For `DynamicArrayVariable` objects, specifying `True` here means the
            name for the underlying data is returned. If specifying `False`,
            the name of object itself is returned (e.g. to allow resizing).

        Returns
        -------
        name : str
            The name for `var`.
        """
        raise NotImplementedError()

    def get_len(self, array):
        """
        Return the length of the array.

        Parameters
        ----------
        array : `ArrayVariable`
            The array for which the length is requested.

        Returns
        -------
        l : int
            The length of the array.
        """
        raise NotImplementedError()

    def add_array(self, var):
        """
        Add an array to this device.

        Parameters
        ----------
        var : `ArrayVariable`
            The array to add.
        """
        raise NotImplementedError()

    def init_with_zeros(self, var, dtype):
        """
        Initialize an array with zeros.

        Parameters
        ----------
        var : `ArrayVariable`
            The array to initialize with zeros.
        dtype : `dtype`
            The data type to use for the array.
        """
        raise NotImplementedError()

    def init_with_arange(self, var, start, dtype):
        """
        Initialize an array with an integer range.

        Parameters
        ----------
        var : `ArrayVariable`
            The array to fill with the integer range.
        start : int
            The start value for the integer range
        dtype : `dtype`
            The data type to use for the array.
        """
        raise NotImplementedError()

    def fill_with_array(self, var, arr):
        """
        Fill an array with the values given in another array.

        Parameters
        ----------
        var : `ArrayVariable`
            The array to fill.
        arr : `ndarray`
            The array values that should be copied to `var`.
        """
        raise NotImplementedError()

    def spike_queue(self, source_start, source_end):
        """
        Create and return a new `SpikeQueue` for this `Device`.

        Parameters
        ----------
        source_start : int
            The start index of the source group (necessary for subgroups)
        source_end : int
            The end index of the source group (necessary for subgroups)
        """
        raise NotImplementedError()

    def resize(self, var, new_size):
        """
        Resize a `DynamicArrayVariable`.

        Parameters
        ----------
        var : `DynamicArrayVariable`
            The variable that should be resized.
        new_size : int
            The new size of the variable
        """
        raise NotImplementedError()

    def resize_along_first(self, var, new_size):
        # Can be overwritten with a better implementation
        return self.resize(var, new_size)

    def seed(self, seed=None):
        """
        Set the seed for the random number generator.

        Parameters
        ----------
        seed : int, optional
            The seed value for the random number generator, or ``None`` (the
            default) to set a random seed.
        """
        raise NotImplementedError()

    def code_object_class(self, codeobj_class=None, fallback_pref="codegen.target"):
        """
        Return `CodeObject` class according to input/default settings

        Parameters
        ----------
        codeobj_class : a `CodeObject` class, optional
            If this is keyword is set to None or no arguments are given, this method will return
            the default.
        fallback_pref : str, optional
            String describing which attribute of prefs to access to retrieve the 'default' target.
            Usually this is codegen.target, but in some cases we want to use object-specific targets
            such as codegen.string_expression_target.

        Returns
        -------
        codeobj_class : class
            The `CodeObject` class that should be used
        """
        if isinstance(codeobj_class, str):
            raise TypeError(
                "codeobj_class argument given to code_object_class device method "
                "should be a CodeObject class, not a string. You can, however, "
                "send a string description of the target desired for the CodeObject "
                "under the keyword fallback_pref"
            )
        if codeobj_class is None:
            codeobj_class = prefs[fallback_pref]
            if isinstance(codeobj_class, str):
                if codeobj_class == "auto":
                    return auto_target()
                for target in codegen_targets:
                    if target.class_name == codeobj_class:
                        return target
                # No target found
                targets = ["auto"] + [
                    target.class_name for target in codegen_targets if target.class_name
                ]
                raise ValueError(
                    f"Unknown code generation target: {codeobj_class}, should be  one"
                    f" of {targets}"
                )
        else:
            return codeobj_class

    def code_object(
        self,
        owner,
        name,
        abstract_code,
        variables,
        template_name,
        variable_indices,
        codeobj_class=None,
        template_kwds=None,
        override_conditional_write=None,
        compiler_kwds=None,
    ):
        if compiler_kwds is None:
            compiler_kwds = {}
        name = find_name(name)
        codeobj_class = self.code_object_class(codeobj_class)
        template = getattr(codeobj_class.templater, template_name)
        iterate_all = template.iterate_all
        generator = codeobj_class.generator_class(
            variables=variables,
            variable_indices=variable_indices,
            owner=owner,
            iterate_all=iterate_all,
            codeobj_class=codeobj_class,
            override_conditional_write=override_conditional_write,
            allows_scalar_write=template.allows_scalar_write,
            name=name,
            template_name=template_name,
        )
        if template_kwds is None:
            template_kwds = dict()
        else:
            template_kwds = template_kwds.copy()

        logger.diagnostic(
            f"{name} abstract code:\n{indent(code_representation(abstract_code))}"
        )

        scalar_code, vector_code, kwds = generator.translate(
            abstract_code, dtype=prefs["core.default_float_dtype"]
        )
        # Add the array names as keywords as well
        for varname, var in variables.items():
            if isinstance(var, ArrayVariable):
                pointer_name = generator.get_array_name(var)
                if var.scalar:
                    pointer_name += "[0]"
                template_kwds[varname] = pointer_name
                if hasattr(var, "resize"):
                    dyn_array_name = generator.get_array_name(var, access_data=False)
                    template_kwds[f"_dynamic_{varname}"] = dyn_array_name

        template_kwds.update(kwds)
        logger.diagnostic(
            f"{name} snippet (scalar):\n{indent(code_representation(scalar_code))}"
        )
        logger.diagnostic(
            f"{name} snippet (vector):\n{indent(code_representation(vector_code))}"
        )

        code = template(
            scalar_code,
            vector_code,
            owner=owner,
            variables=variables,
            codeobj_name=name,
            variable_indices=variable_indices,
            get_array_name=generator.get_array_name,
            **template_kwds,
        )
        logger.diagnostic(f"{name} code:\n{indent(code_representation(code))}")

        codeobj = codeobj_class(
            owner,
            code,
            variables,
            variable_indices,
            template_name=template_name,
            template_source=template.template_source,
            name=name,
            compiler_kwds=compiler_kwds,
        )
        codeobj.compile()
        return codeobj

    def activate(self, build_on_run=True, **kwargs):
        """
        Called when this device is set as the current device.
        """
        from brian2.core.clocks import Clock  # avoid import issues

        if self.defaultclock is None:
            self.defaultclock = Clock(dt=0.1 * ms, name="defaultclock")
        self._set_maximum_run_time(None)
        self.build_on_run = build_on_run
        self.build_options = dict(kwargs)

    def insert_device_code(self, slot, code):
        # Deprecated
        raise AttributeError(
            "The method 'insert_device_code' has been renamed to 'insert_code'."
        )

    def insert_code(self, slot, code):
        """
        Insert code directly into a given slot in the device. By default does nothing.
        """
        logger.warn(f"Ignoring device code, unknown slot: {slot}, code: {code}")

    def build(self, **kwds):
        """
        For standalone projects, called when the project is ready to be built. Does nothing for runtime mode.
        """
        pass

    def reinit(self):
        """
        Reinitialize the device. For standalone devices, clears all the internal
        state of the device.
        """
        pass

    def delete(self, data=True, code=True, directory=True, force=False):
        """
        Delete code and/or data generated/stored by the device.

        Parameters
        ----------
        data : bool, optional
            Whether to delete the data generated by the simulation (final values
            of state variables, data stored in monitors, etc.). Defaults to
            ``True``.
        code : bool, optional
            Whether to delete the code generated by the simulation. Includes the
            numerical values used for initialization of state variables in
            assignments not using strings. Defaults to ``True``.
        directory : bool, optional
            Whether to delete the project directory generated by the simulation.
            Will not delete directories that contain files not created by Brian
            unless the ``force`` option is specfied. Defaults to ``True``.
        force : bool, optional
            Whether to delete the project directory with all its content, even
            if it contains files that were not created by Brian. Useful only
            when the ``directory`` option is set to ``True`` as well. Defaults
            to ``False``.
        """
        pass

    def get_random_state(self):
        """
        Return a (pickable) representation of the current random number
        generator state. Providing the returned object (e.g. a dict) to
        `.Device.set_random_state` should restore the random number generator
        state.

        Returns
        -------
        state
            The state of the random number generator in a representation
            that can be passed as an argument to `.Device.set_random_state`.
        """
        raise NotImplementedError(
            "Device does not support getting the state of the random number generator."
        )

    def set_random_state(self, state):
        """
        Reset the random number generator state to a previously stored state
        (see `.Device.get_random_state`).

        Parameters
        ----------
        state
            A random number generator state as provided by
            `Device.get_random_state`.
        """
        raise NotImplementedError(
            "Device does not support setting the state of the random number generator."
        )


class RuntimeDevice(Device):
    """
    The default device used in Brian, state variables are stored as numpy
    arrays in memory.
    """

    def __init__(self):
        super().__init__()
        #: Mapping from `Variable` objects to numpy arrays (or `DynamicArray`
        #: objects). Arrays in this dictionary will disappear as soon as the
        #: last reference to the `Variable` object used as a key is gone
        self.arrays = WeakKeyDictionary()
        # Note that the buffers only store a pointer to the actual random
        # numbers -- the buffer will be filled in Cython code
        self.randn_buffer = np.zeros(1, dtype=np.intp)
        self.rand_buffer = np.zeros(1, dtype=np.intp)
        self.randn_buffer_index = np.zeros(1, dtype=np.int32)
        self.rand_buffer_index = np.zeros(1, dtype=np.int32)

    def __getstate__(self):
        state = dict(self.__dict__)
        # Python's pickle module cannot pickle a WeakKeyDictionary, we therefore
        # convert it to a standard dictionary
        state["arrays"] = dict(self.arrays)
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self.__dict__["arrays"] = WeakKeyDictionary(self.__dict__["arrays"])

    def get_array_name(self, var, access_data=True):
        # if no owner is set, this is a temporary object (e.g. the array
        # of indices when doing G.x[indices] = ...). The name is not
        # necessarily unique over several CodeObjects in this case.
        owner_name = getattr(var.owner, "name", "temporary")

        if isinstance(var, DynamicArrayVariable):
            if access_data:
                return f"_array_{owner_name}_{var.name}"
            else:
                return f"_dynamic_array_{owner_name}_{var.name}"
        elif isinstance(var, ArrayVariable):
            return f"_array_{owner_name}_{var.name}"
        else:
            raise TypeError(f"Do not have a name for variable of type {type(var)}.")

    def add_array(self, var):
        # This creates the actual numpy arrays (or DynamicArrayVariable objects)
        if isinstance(var, DynamicArrayVariable):
            if var.ndim == 1:
                arr = DynamicArray1D(var.size, dtype=var.dtype)
            else:
                arr = DynamicArray(var.size, dtype=var.dtype)
        else:
            arr = np.empty(var.size, dtype=var.dtype)

        self.arrays[var] = arr

    def get_value(self, var, access_data=True):
        if isinstance(var, DynamicArrayVariable) and access_data:
            return self.arrays[var].data
        else:
            return self.arrays[var]

    def set_value(self, var, value):
        self.arrays[var][:] = value

    def resize(self, var, new_size):
        self.arrays[var].resize(new_size)

    def resize_along_first(self, var, new_size):
        self.arrays[var].resize_along_first(new_size)

    def init_with_zeros(self, var, dtype):
        self.arrays[var][:] = 0

    def init_with_arange(self, var, start, dtype):
        self.arrays[var][:] = np.arange(start, stop=var.get_len() + start, dtype=dtype)

    def fill_with_array(self, var, arr):
        self.arrays[var][:] = arr

    def spike_queue(self, source_start, source_end):
        # Use the C++ version of the SpikeQueue when available
        try:
            from brian2.synapses.cythonspikequeue import SpikeQueue

            logger.diagnostic("Using the C++ SpikeQueue", once=True)
        except ImportError:
            from brian2.synapses.spikequeue import SpikeQueue

            logger.diagnostic("Using the Python SpikeQueue", once=True)

        return SpikeQueue(source_start=source_start, source_end=source_end)

    def seed(self, seed=None):
        """
        Set the seed for the random number generator.

        Parameters
        ----------
        seed : int, optional
            The seed value for the random number generator, or ``None`` (the
            default) to set a random seed.
        """
        np.random.seed(seed)
        self.rand_buffer_index[:] = 0
        self.randn_buffer_index[:] = 0

    def get_random_state(self):
        return {
            "numpy_state": np.random.get_state(),
            "rand_buffer_index": np.array(self.rand_buffer_index),
            "rand_buffer": np.array(self.rand_buffer),
            "randn_buffer_index": np.array(self.randn_buffer_index),
            "randn_buffer": np.array(self.randn_buffer),
        }

    def set_random_state(self, state):
        np.random.set_state(state["numpy_state"])
        self.rand_buffer_index[:] = state["rand_buffer_index"]
        self.rand_buffer[:] = state["rand_buffer"]
        self.randn_buffer_index[:] = state["randn_buffer_index"]
        self.randn_buffer[:] = state["randn_buffer"]


class Dummy:
    """
    Dummy object
    """

    def __getattr__(self, name):
        return Dummy()

    def __call__(self, *args, **kwds):
        return Dummy()

    def __enter__(self):
        return Dummy()

    def __exit__(self, type, value, traceback):
        pass

    def __getitem__(self, i):
        return Dummy()

    def __setitem__(self, i, val):
        pass


class CurrentDeviceProxy:
    """
    Method proxy for access to the currently active device
    """

    def __getattr__(self, name):
        if not hasattr(active_device, name):
            # We special case the name "shape" here, since some IDEs (e.g. The Python
            # console in PyDev and PyCharm) use the "shape" attribute to determine
            # whether an object is "array-like".
            if name.startswith("_") or name == "shape":
                # Do not fake private/magic attributes
                raise AttributeError(
                    f"Active device does not have an attribute '{name}'."
                )
            else:
                logger.warn(
                    f"Active device does not have an attribute '{name}', ignoring this."
                )
                attr = Dummy()
        else:
            attr = getattr(active_device, name)
        return attr


#: Proxy object to access methods of the current device
device = CurrentDeviceProxy()
#: The currently active device (set with `set_device`)
active_device = None


def get_device():
    """
    Gets the actve `Device` object
    """
    global active_device
    return active_device


#: A stack of previously set devices as a tuple with their options (see
#: `set_device`): (device, build_on_run, build_options)
previous_devices = []


def set_device(device, build_on_run=True, **kwargs):
    """
    Set the device used for simulations.

    Parameters
    ----------
    device : `Device` or str
        The `Device` object or the name of the device.
    build_on_run : bool, optional
        Whether a call to `run` (or `Network.run`) should directly trigger a
        `Device.build`. This is only relevant for standalone devices and means
        that a run call directly triggers the start of a simulation. If the
        simulation consists of multiple run calls, set ``build_on_run`` to
        ``False`` and call `Device.build` explicitly. Defaults to ``True``.
    kwargs : dict, optional
        Only relevant when ``build_on_run`` is ``True``: additional arguments
        that will be given to the `Device.build` call.
    """
    global previous_devices
    if active_device is not None:
        prev_build_on_run = getattr(active_device, "build_on_run", True)
        prev_build_options = getattr(active_device, "build_options", {})
        previous_devices.append((active_device, prev_build_on_run, prev_build_options))
    _do_set_device(device, build_on_run, **kwargs)


def _do_set_device(device, build_on_run=True, **kwargs):
    global active_device

    if isinstance(device, str):
        device = all_devices[device]
    if active_device is not None and active_device.defaultclock is not None:
        previous_dt = active_device.defaultclock.dt
    else:
        previous_dt = None
    active_device = device

    active_device.activate(build_on_run=build_on_run, **kwargs)
    if previous_dt is not None:
        # Copy over the dt information of the defaultclock
        active_device.defaultclock.dt = previous_dt


def reset_device(device=None):
    """
    Reset to a previously used device. Restores also the previously specified
    build options (see `set_device`) for the device. Mostly useful for internal
    Brian code and testing on various devices.

    Parameters
    ----------
    device : `Device` or str, optional
        The device to go back to. If none is specified, go back to the device
        chosen with `set_device` before the current one.
    """
    global previous_devices
    if isinstance(device, str):
        device = all_devices[device]

    if len(previous_devices) == 0 and device is None:
        device = runtime_device
        build_on_run = True
        build_options = {}
    elif device is None:
        device, build_on_run, build_options = previous_devices.pop()
    else:
        build_on_run = device.build_on_run
        build_options = device.build_options

    _do_set_device(device, build_on_run, **build_options)


def reinit_devices():
    """
    Reinitialize all devices, call `Device.activate` again on the current
    device and reset the preferences. Used as a "teardown" function in testing,
    if users want to reset their device (e.g. for multiple standalone runs in a
    single script), calling ``device.reinit()`` followed by
    ``device.activate()`` should normally be sufficient.

    Notes
    -----
    This also resets the `defaultclock`, i.e. a non-standard ``dt`` has to be
    set again.
    """
    from brian2 import restore_initial_state  # avoids circular import

    for device in all_devices.values():
        device.reinit()

    if active_device is not None:
        # Reactivate the current device
        reset_device(active_device)

    restore_initial_state()


def reinit_and_delete():
    """
    Calls `reinit_devices` and additionally deletes the files left behind by
    the standalone mode in the temporary directory.
    Silently suppresses errors that occur while deleting the directory.
    """
    reinit_devices()
    device.delete(directory=True, force=True)


def seed(seed=None):
    """
    Set the seed for the random number generator.

    Parameters
    ----------
    seed : int, optional
        The seed value for the random number generator, or ``None`` (the
        default) to set a random seed.

    Notes
    -----
    This function delegates the call to `Device.seed` of the current device.
    """
    if seed is not None and not isinstance(seed, numbers.Integral):
        raise TypeError(f"Seed has to be None or an integer, was {type(seed)}")
    get_device().seed(seed)


runtime_device = RuntimeDevice()
all_devices["runtime"] = runtime_device



"""
Module defining `EventMonitor` and `SpikeMonitor`.
"""

import numpy as np

from brian2.core.names import Nameable
from brian2.core.spikesource import SpikeSource
from brian2.core.variables import Variables
from brian2.groups.group import CodeRunner, Group
from brian2.units.fundamentalunits import Quantity

__all__ = ["EventMonitor", "SpikeMonitor"]


class EventMonitor(Group, CodeRunner):
    """
    Record events from a `NeuronGroup` or another event source.

    The recorded events can be accessed in various ways:
    the attributes `~EventMonitor.i` and `~EventMonitor.t` store all the indices
    and event times, respectively. Alternatively, you can get a dictionary
    mapping neuron indices to event trains, by calling the `event_trains`
    method.

    Parameters
    ----------
    source : `NeuronGroup`, `SpikeSource`
        The source of events to record.
    event : str
        The name of the event to record
    variables : str or sequence of str, optional
        Which variables to record at the time of the event (in addition to the
        index of the neuron). Can be the name of a variable or a list of names.
    record : bool, optional
        Whether or not to record each event in `i` and `t` (the `count` will
        always be recorded). Defaults to ``True``.
    when : str, optional
        When to record the events, by default records events in the same slot
        where the event is emitted. See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to the order where the
        event is emitted + 1, i.e. it will be recorded directly afterwards.
    name : str, optional
        A unique name for the object, otherwise will use
        ``source.name+'_eventmonitor_0'``, etc.
    codeobj_class : class, optional
        The `CodeObject` class to run code with.

    See Also
    --------
    SpikeMonitor
    """

    invalidates_magic_network = False
    add_to_magic_network = True

    def __init__(
        self,
        source,
        event,
        variables=None,
        record=True,
        when=None,
        order=None,
        name="eventmonitor*",
        codeobj_class=None,
    ):
        if not isinstance(source, SpikeSource):
            raise TypeError(
                f"{self.__class__.__name__} can only monitor groups "
                "producing spikes (such as NeuronGroup), but the given "
                f"argument is of type {type(source)}."
            )
        #: The source we are recording from
        self.source = source
        #: Whether to record times and indices of events
        self.record = record
        #: The array of event counts (length = size of target group)
        self.count = None
        del self.count  # this is handled by the Variable mechanism

        if event not in source.events:
            if event == "spike":
                threshold_text = " Did you forget to set a 'threshold'?"
            else:
                threshold_text = ""
            raise ValueError(
                f"Recorded group '{source.name}' does not define an event "
                f"'{event}'.{threshold_text}"
            )
        if when is None:
            if order is not None:
                raise ValueError("Cannot specify order if when is not specified.")
            # TODO: Would be nicer if there was a common way of accessing the
            #       relevant object for NeuronGroup and SpikeGeneratorGroup
            if hasattr(source, "thresholder"):
                parent_obj = source.thresholder[event]
            else:
                parent_obj = source
            when = parent_obj.when
            order = parent_obj.order + 1
        elif order is None:
            order = 0

        #: The event that we are listening to
        self.event = event

        if variables is None:
            variables = {}
        elif isinstance(variables, str):
            variables = {variables}

        #: The additional variables that will be recorded
        self.record_variables = set(variables)

        for variable in variables:
            if variable not in source.variables:
                raise ValueError(
                    f"'{variable}' is not a variable of the recorded group"
                )

        if self.record:
            self.record_variables |= {"i", "t"}

        # Some dummy code so that code generation takes care of the indexing
        # and subexpressions
        code = [f"_to_record_{v} = _source_{v}" for v in sorted(self.record_variables)]
        code = "\n".join(code)

        self.codeobj_class = codeobj_class

        # Since this now works for general events not only spikes, we have to
        # pass the information about which variable to use to the template,
        # it can not longer simply refer to "_spikespace"
        eventspace_name = f"_{event}space"

        # Handle subgroups correctly
        start = getattr(source, "start", 0)
        stop = getattr(source, "stop", len(source))
        source_N = getattr(source, "_source_N", len(source))

        Nameable.__init__(self, name=name)

        self.variables = Variables(self)
        self.variables.add_reference(eventspace_name, source)

        for variable in self.record_variables:
            source_var = source.variables[variable]
            self.variables.add_reference(f"_source_{variable}", source, variable)
            self.variables.add_auxiliary_variable(
                f"_to_record_{variable}",
                dimensions=source_var.dim,
                dtype=source_var.dtype,
            )
            self.variables.add_dynamic_array(
                variable,
                size=0,
                dimensions=source_var.dim,
                dtype=source_var.dtype,
                read_only=True,
            )
        self.variables.add_arange("_source_idx", size=len(source))
        self.variables.add_array(
            "count",
            size=len(source),
            dtype=np.int32,
            read_only=True,
            index="_source_idx",
        )
        self.variables.add_constant("_source_start", start)
        self.variables.add_constant("_source_stop", stop)
        self.variables.add_constant("_source_N", source_N)
        self.variables.add_array(
            "N", size=1, dtype=np.int32, read_only=True, scalar=True
        )

        record_variables = {
            varname: self.variables[varname] for varname in self.record_variables
        }
        template_kwds = {
            "eventspace_variable": source.variables[eventspace_name],
            "record_variables": record_variables,
            "record": self.record,
        }
        needed_variables = {eventspace_name} | self.record_variables
        CodeRunner.__init__(
            self,
            group=self,
            code=code,
            template="spikemonitor",
            name=None,  # The name has already been initialized
            clock=source.clock,
            when=when,
            order=order,
            needed_variables=needed_variables,
            template_kwds=template_kwds,
        )

        self.variables.create_clock_variables(self._clock, prefix="_clock_")
        self.add_dependency(source)
        self.written_readonly_vars = {
            self.variables[varname] for varname in self.record_variables
        }
        self._enable_group_attributes()

    def resize(self, new_size):
        # Note that this does not set N, this has to be done in the template
        # since we use a restricted pointer to access it (which promises that
        # we only change the value through this pointer)
        for variable in self.record_variables:
            self.variables[variable].resize(new_size)

    def reinit(self):
        """
        Clears all recorded spikes
        """
        raise NotImplementedError()

    @property
    def it(self):
        """
        Returns the pair (`i`, `t`).
        """
        if not self.record:
            raise AttributeError(
                "Indices and times have not been recorded."
                "Set the record argument to True to record "
                "them."
            )
        return self.i, self.t

    @property
    def it_(self):
        """
        Returns the pair (`i`, `t_`).
        """
        if not self.record:
            raise AttributeError(
                "Indices and times have not been recorded."
                "Set the record argument to True to record "
                "them."
            )

        return self.i, self.t_

    def _values_dict(self, first_pos, sort_indices, used_indices, var):
        sorted_values = self.state(var, use_units=False)[sort_indices]
        dim = self.variables[var].dim
        event_values = {}
        current_pos = 0  # position in the all_indices array
        for idx in range(len(self.source)):
            if current_pos < len(used_indices) and used_indices[current_pos] == idx:
                if current_pos < len(used_indices) - 1:
                    event_values[idx] = Quantity(
                        sorted_values[
                            first_pos[current_pos] : first_pos[current_pos + 1]
                        ],
                        dim=dim,
                    )
                else:
                    event_values[idx] = Quantity(
                        sorted_values[first_pos[current_pos] :], dim=dim
                    )
                current_pos += 1
            else:
                event_values[idx] = Quantity([], dim=dim)
        return event_values

    def values(self, var):
        """
        Return a dictionary mapping neuron indices to arrays of variable values
        at the time of the events (sorted by time).

        Parameters
        ----------
        var : str
            The name of the variable.

        Returns
        -------
        values : dict
            Dictionary mapping each neuron index to an array of variable
            values at the time of the events

        Examples
        --------
        >>> from brian2 import *
        >>> G = NeuronGroup(2, '''counter1 : integer
        ...                       counter2 : integer
        ...                       max_value : integer''',
        ...                    threshold='counter1 >= max_value',
        ...                    reset='counter1 = 0')
        >>> G.run_regularly('counter1 += 1; counter2 += 1')  # doctest: +ELLIPSIS
        CodeRunner(...)
        >>> G.max_value = [50, 100]
        >>> mon = EventMonitor(G, event='spike', variables='counter2')
        >>> run(10*ms)
        >>> counter2_values = mon.values('counter2')
        >>> print(counter2_values[0])
        [ 50 100]
        >>> print(counter2_values[1])
        [100]
        """
        if not self.record:
            raise AttributeError(
                "Indices and times have not been recorded."
                "Set the record argument to True to record "
                "them."
            )
        indices = self.i[:]
        # We have to make sure that the sort is stable, otherwise our spike
        # times do not necessarily remain sorted.
        sort_indices = np.argsort(indices, kind="mergesort")
        used_indices, first_pos = np.unique(self.i[:][sort_indices], return_index=True)
        return self._values_dict(first_pos, sort_indices, used_indices, var)

    def all_values(self):
        """
        Return a dictionary mapping recorded variable names (including ``t``)
        to a dictionary mapping neuron indices to arrays of variable values at
        the time of the events (sorted by time). This is equivalent to (but more
        efficient than) calling `values` for each variable and storing the
        result in a dictionary.

        Returns
        -------
        all_values : dict
            Dictionary mapping variable names to dictionaries which themselves
            are mapping neuron indicies to arrays of variable values at the
            time of the events.

        Examples
        --------
        >>> from brian2 import *
        >>> G = NeuronGroup(2, '''counter1 : integer
        ...                       counter2 : integer
        ...                       max_value : integer''',
        ...                    threshold='counter1 >= max_value',
        ...                    reset='counter1 = 0')
        >>> G.run_regularly('counter1 += 1; counter2 += 1')  # doctest: +ELLIPSIS
        CodeRunner(...)
        >>> G.max_value = [50, 100]
        >>> mon = EventMonitor(G, event='spike', variables='counter2')
        >>> run(10*ms)
        >>> all_values = mon.all_values()
        >>> print(all_values['counter2'][0])
        [ 50 100]
        >>> print(all_values['t'][1])
        [ 9.9] ms
        """
        if not self.record:
            raise AttributeError(
                "Indices and times have not been recorded."
                "Set the record argument to True to record "
                "them."
            )
        indices = self.i[:]
        sort_indices = np.argsort(indices, kind="mergesort")
        used_indices, first_pos = np.unique(self.i[:][sort_indices], return_index=True)
        all_values_dict = {}
        for varname in self.record_variables - {"i"}:
            all_values_dict[varname] = self._values_dict(
                first_pos, sort_indices, used_indices, varname
            )
        return all_values_dict

    def event_trains(self):
        """
        Return a dictionary mapping neuron indices to arrays of event times.
        Equivalent to calling ``values('t')``.

        Returns
        -------
        event_trains : dict
            Dictionary that stores an array with the event times for each
            neuron index.

        See Also
        --------
        SpikeMonitor.spike_trains
        """
        return self.values("t")

    @property
    def num_events(self):
        """
        Returns the total number of recorded events.
        """
        return self.N[:]

    def __repr__(self):
        classname = self.__class__.__name__
        return f"<{classname}, recording event '{self.event}' from '{self.group.name}'>"


class SpikeMonitor(EventMonitor):
    """
    Record spikes from a `NeuronGroup` or other spike source.

    The recorded spikes can be accessed in various ways (see Examples below):
    the attributes `~SpikeMonitor.i` and `~SpikeMonitor.t` store all the indices
    and spike times, respectively. Alternatively, you can get a dictionary
    mapping neuron indices to spike trains, by calling the `spike_trains`
    method. If you record additional variables with the ``variables`` argument,
    these variables can be accessed by their name (see Examples).

    Parameters
    ----------
    source : (`NeuronGroup`, `SpikeSource`)
        The source of spikes to record.
    variables : str or sequence of str, optional
        Which variables to record at the time of the spike (in addition to the
        index of the neuron). Can be the name of a variable or a list of names.
    record : bool, optional
        Whether or not to record each spike in `i` and `t` (the `count` will
        always be recorded). Defaults to ``True``.
    when : str, optional
        When to record the events, by default records events in the same slot
        where the event is emitted. See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to the order where the
        event is emitted + 1, i.e. it will be recorded directly afterwards.
    name : str, optional
        A unique name for the object, otherwise will use
        ``source.name+'_spikemonitor_0'``, etc.
    codeobj_class : class, optional
        The `CodeObject` class to run code with.

    Examples
    --------
    >>> from brian2 import *
    >>> spikes = SpikeGeneratorGroup(3, [0, 1, 2], [0, 1, 2]*ms)
    >>> spike_mon = SpikeMonitor(spikes)
    >>> net = Network(spikes, spike_mon)
    >>> net.run(3*ms)
    >>> print(spike_mon.i[:])
    [0 1 2]
    >>> print(spike_mon.t[:])
    [ 0.  1.  2.] ms
    >>> print(spike_mon.t_[:])
    [ 0.     0.001  0.002]
    >>> from brian2 import *
    >>> G = NeuronGroup(2, '''counter1 : integer
    ...                       counter2 : integer
    ...                       max_value : integer''',
    ...                    threshold='counter1 >= max_value',
    ...                    reset='counter1 = 0')
    >>> G.run_regularly('counter1 += 1; counter2 += 1')  # doctest: +ELLIPSIS
    CodeRunner(...)
    >>> G.max_value = [50, 100]
    >>> mon = SpikeMonitor(G, variables='counter2')
    >>> net = Network(G, mon)
    >>> net.run(10*ms)
    >>> print(mon.i[:])
    [0 0 1]
    >>> print(mon.counter2[:])
    [ 50 100 100]
    """

    def __init__(
        self,
        source,
        variables=None,
        record=True,
        when=None,
        order=None,
        name="spikemonitor*",
        codeobj_class=None,
    ):
        #: The array of spike counts (length = size of target group)
        self.count = None
        del self.count  # this is handled by the Variable mechanism
        super().__init__(
            source,
            event="spike",
            variables=variables,
            record=record,
            when=when,
            order=order,
            name=name,
            codeobj_class=codeobj_class,
        )

    @property
    def num_spikes(self):
        """
        Returns the total number of recorded spikes.
        """
        return self.num_events

    # We "re-implement" the following functions only to get more specific
    # doc strings (and to make sure that the methods are included in the
    # reference documentation for SpikeMonitor).

    def spike_trains(self):
        """
        Return a dictionary mapping neuron indices to arrays of spike times.

        Returns
        -------
        spike_trains : dict
            Dictionary that stores an array with the spike times for each
            neuron index.

        Examples
        --------
        >>> from brian2 import *
        >>> spikes = SpikeGeneratorGroup(3, [0, 1, 2], [0, 1, 2]*ms)
        >>> spike_mon = SpikeMonitor(spikes)
        >>> run(3*ms)
        >>> spike_trains = spike_mon.spike_trains()
        >>> spike_trains[1]
        array([ 1.]) * msecond
        """
        return self.event_trains()

    def values(self, var):
        """
        Return a dictionary mapping neuron indices to arrays of variable values
        at the time of the spikes (sorted by time).

        Parameters
        ----------
        var : str
            The name of the variable.

        Returns
        -------
        values : dict
            Dictionary mapping each neuron index to an array of variable
            values at the time of the spikes.

        Examples
        --------
        >>> from brian2 import *
        >>> G = NeuronGroup(2, '''counter1 : integer
        ...                       counter2 : integer
        ...                       max_value : integer''',
        ...                    threshold='counter1 >= max_value',
        ...                    reset='counter1 = 0')
        >>> G.run_regularly('counter1 += 1; counter2 += 1')  # doctest: +ELLIPSIS
        CodeRunner(...)
        >>> G.max_value = [50, 100]
        >>> mon = SpikeMonitor(G, variables='counter2')
        >>> run(10*ms)
        >>> counter2_values = mon.values('counter2')
        >>> print(counter2_values[0])
        [ 50 100]
        >>> print(counter2_values[1])
        [100]
        """
        return super().values(var)

    def all_values(self):
        """
        Return a dictionary mapping recorded variable names (including ``t``)
        to a dictionary mapping neuron indices to arrays of variable values at
        the time of the spikes (sorted by time). This is equivalent to (but more
        efficient than) calling `values` for each variable and storing the
        result in a dictionary.

        Returns
        -------
        all_values : dict
            Dictionary mapping variable names to dictionaries which themselves
            are mapping neuron indicies to arrays of variable values at the
            time of the spikes.

        Examples
        --------
        >>> from brian2 import *
        >>> G = NeuronGroup(2, '''counter1 : integer
        ...                       counter2 : integer
        ...                       max_value : integer''',
        ...                    threshold='counter1 >= max_value',
        ...                    reset='counter1 = 0')
        >>> G.run_regularly('counter1 += 1; counter2 += 1')  # doctest: +ELLIPSIS
        CodeRunner(...)
        >>> G.max_value = [50, 100]
        >>> mon = SpikeMonitor(G, variables='counter2')
        >>> run(10*ms)
        >>> all_values = mon.all_values()
        >>> print(all_values['counter2'][0])
        [ 50 100]
        >>> print(all_values['t'][1])
        [ 9.9] ms
        """
        return super().all_values()

    def __repr__(self):
        classname = self.__class__.__name__
        return f"<{classname}, recording from '{self.group.name}'>"



from brian2 import *
from brian2.tests.features import *
from brian2.tests.features.base import *
from brian2.tests.features.speed import *
#from brian2genn.correctness_testing import GeNNConfiguration
import os, pickle

use_cached_results = True

if use_cached_results and os.path.exists('cached_speed_test_results.pkl'):
    with open('cached_speed_test_results.pkl', 'rb') as f:
        res = pickle.load(f)
else:
    # Full testing
    # res = run_speed_tests()

    # Quick testing
    res = run_speed_tests(configurations=[NumpyConfiguration,
                                          CythonConfiguration,
                                          #LocalConfiguration,
                                          CPPStandaloneConfiguration,
                                          #CPPStandaloneConfigurationOpenMP,
    #                                      GeNNConfiguration,
                                          ],
                          speed_tests=[
                                       LinearNeuronsOnly,
                                       HHNeuronsOnly,
                                       CUBAFixedConnectivity,
                                       COBAHHFixedConnectivity,
                                       VerySparseMediumRateSynapsesOnly,
                                       SparseMediumRateSynapsesOnly,
                                       DenseMediumRateSynapsesOnly,
                                       SparseLowRateSynapsesOnly,
                                       SparseHighRateSynapsesOnly,
                                       STDP,
                                       ],
                          #n_slice=slice(None, None, 3),
                          # n_slice=slice(None, -3),
                          run_twice=False,
                          maximum_run_time=1*second,
                          )

    with open('cached_speed_test_results.pkl', 'wb') as f:
        pickle.dump(res, f, -1)

res.plot_all_tests()
# res.plot_all_tests(profiling_minimum=0.15)
res.plot_all_tests(relative=True)
show()



"""
Numerical integration functions.
"""

import operator
import string
from functools import reduce

import sympy
from pyparsing import (
    Group,
    Literal,
    ParseException,
    Suppress,
    Word,
    ZeroOrMore,
    restOfLine,
)
from sympy.core.sympify import SympifyError

from brian2.equations.codestrings import is_constant_over_dt
from brian2.parsing.sympytools import str_to_sympy, sympy_to_str

from .base import (
    StateUpdateMethod,
    UnsupportedEquationsException,
    extract_method_options,
)

__all__ = ["milstein", "heun", "euler", "rk2", "rk4", "ExplicitStateUpdater"]


# ===============================================================================
# Class for simple definition of explicit state updaters
# ===============================================================================


def _symbol(name, positive=None):
    """Shorthand for ``sympy.Symbol(name, real=True)``."""
    return sympy.Symbol(name, real=True, positive=positive)


#: reserved standard symbols
SYMBOLS = {
    "__x": _symbol("__x"),
    "__t": _symbol("__t", positive=True),
    "dt": _symbol("dt", positive=True),
    "t": _symbol("t", positive=True),
    "__f": sympy.Function("__f"),
    "__g": sympy.Function("__g"),
    "__dW": _symbol("__dW"),
}


def split_expression(expr):
    """
    Split an expression into a part containing the function ``f`` and another
    one containing the function ``g``. Returns a tuple of the two expressions
    (as sympy expressions).

    Parameters
    ----------
    expr : str
        An expression containing references to functions ``f`` and ``g``.

    Returns
    -------
    (non_stochastic, stochastic) : tuple of sympy expressions
        A pair of expressions representing the non-stochastic (containing
        function-independent terms and terms involving ``f``) and the
        stochastic part of the expression (terms involving ``g`` and/or ``dW``).

    Examples
    --------
    >>> split_expression('dt * __f(__x, __t)')
    (dt*__f(__x, __t), None)
    >>> split_expression('dt * __f(__x, __t) + __dW * __g(__x, __t)')
    (dt*__f(__x, __t), __dW*__g(__x, __t))
    >>> split_expression('1/(2*sqrt(dt))*(__g_support - __g(__x, __t))*(sqrt(__dW))')
    (0, sqrt(__dW)*__g_support/(2*sqrt(dt)) - sqrt(__dW)*__g(__x, __t)/(2*sqrt(dt)))
    """

    f = SYMBOLS["__f"]
    g = SYMBOLS["__g"]
    dW = SYMBOLS["__dW"]
    # Arguments of the f and g functions
    x_f = sympy.Wild("x_f", exclude=[f, g], real=True)
    t_f = sympy.Wild("t_f", exclude=[f, g], real=True)
    x_g = sympy.Wild("x_g", exclude=[f, g], real=True)
    t_g = sympy.Wild("t_g", exclude=[f, g], real=True)

    # Reorder the expression so that f(x,t) and g(x,t) are factored out
    sympy_expr = sympy.sympify(expr, locals=SYMBOLS).expand()
    sympy_expr = sympy.collect(sympy_expr, f(x_f, t_f))
    sympy_expr = sympy.collect(sympy_expr, g(x_g, t_g))

    # Constant part, contains neither f, g nor dW
    independent = sympy.Wild("independent", exclude=[f, g, dW], real=True)
    # The exponent of the random number
    dW_exponent = sympy.Wild("dW_exponent", exclude=[f, g, dW, 0], real=True)
    # The factor for the random number, not containing the g function
    independent_dW = sympy.Wild("independent_dW", exclude=[f, g, dW], real=True)
    # The factor for the f function
    f_factor = sympy.Wild("f_factor", exclude=[f, g], real=True)
    # The factor for the g function
    g_factor = sympy.Wild("g_factor", exclude=[f, g], real=True)

    match_expr = (
        independent
        + f_factor * f(x_f, t_f)
        + independent_dW * dW**dW_exponent
        + g_factor * g(x_g, t_g)
    )
    matches = sympy_expr.match(match_expr)

    if matches is None:
        raise ValueError(
            f'Expression "{sympy_expr}" in the state updater description could not be'
            " parsed."
        )

    # Non-stochastic part
    if x_f in matches:
        # Includes the f function
        non_stochastic = matches[independent] + (
            matches[f_factor] * f(matches[x_f], matches[t_f])
        )
    else:
        # Does not include f, might be 0
        non_stochastic = matches[independent]

    # Stochastic part
    if independent_dW in matches and matches[independent_dW] != 0:
        # includes a random variable term with a non-zero factor
        stochastic = (
            matches[g_factor] * g(matches[x_g], matches[t_g])
            + matches[independent_dW] * dW ** matches[dW_exponent]
        )
    elif x_g in matches:
        # Does not include a random variable but the g function
        stochastic = matches[g_factor] * g(matches[x_g], matches[t_g])
    else:
        # Contains neither random variable nor g function --> empty
        stochastic = None

    return (non_stochastic, stochastic)


class ExplicitStateUpdater(StateUpdateMethod):
    """
    An object that can be used for defining state updaters via a simple
    description (see below). Resulting instances can be passed to the
    ``method`` argument of the `NeuronGroup` constructor. As other state
    updater functions the `ExplicitStateUpdater` objects are callable,
    returning abstract code when called with an `Equations` object.

    A description of an explicit state updater consists of a (multi-line)
    string, containing assignments to variables and a final "x_new = ...",
    stating the integration result for a single timestep. The assignments
    can be used to define an arbitrary number of intermediate results and
    can refer to ``f(x, t)`` (the function being integrated, as a function of
    ``x``, the previous value of the state variable and ``t``, the time) and
    ``dt``, the size of the timestep.

    For example, to define a Runge-Kutta 4 integrator (already provided as
    `rk4`), use::

            k1 = dt*f(x,t)
            k2 = dt*f(x+k1/2,t+dt/2)
            k3 = dt*f(x+k2/2,t+dt/2)
            k4 = dt*f(x+k3,t+dt)
            x_new = x+(k1+2*k2+2*k3+k4)/6

    Note that for stochastic equations, the function `f` only corresponds to
    the non-stochastic part of the equation. The additional function `g`
    corresponds to the stochastic part that has to be multiplied with the
    stochastic variable xi (a standard normal random variable -- if the
    algorithm needs a random variable with a different variance/mean you have
    to multiply/add it accordingly). Equations with more than one
    stochastic variable do not have to be treated differently, the part
    referring to ``g`` is repeated for all stochastic variables automatically.

    Stochastic integrators can also make reference to ``dW`` (a normal
    distributed random number with variance ``dt``) and ``g(x, t)``, the
    stochastic part of an equation. A stochastic state updater could therefore
    use a description like::

        x_new = x + dt*f(x,t) + g(x, t) * dW

    For simplicity, the same syntax is used for state updaters that only support
    additive noise, even though ``g(x, t)`` does not depend on ``x`` or ``t``
    in that case.

    There a some restrictions on the complexity of the expressions (but most
    can be worked around by using intermediate results as in the above Runge-
    Kutta example): Every statement can only contain the functions ``f`` and
    ``g`` once; The expressions have to be linear in the functions, e.g. you
    can use ``dt*f(x, t)`` but not ``f(x, t)**2``.

    Parameters
    ----------
    description : str
        A state updater description (see above).
    stochastic : {None, 'additive', 'multiplicative'}
        What kind of stochastic equations this state updater supports: ``None``
        means no support of stochastic equations, ``'additive'`` means only
        equations with additive noise and ``'multiplicative'`` means
        supporting arbitrary stochastic equations.

    Raises
    ------
    ValueError
        If the parsing of the description failed.

    Notes
    -----
    Since clocks are updated *after* the state update, the time ``t`` used
    in the state update step is still at its previous value. Enumerating the
    states and discrete times, ``x_new = x + dt*f(x, t)`` is therefore
    understood as :math:`x_{i+1} = x_i + dt f(x_i, t_i)`, yielding the correct
    forward Euler integration. If the integrator has to refer to the time at
    the end of the timestep, simply use ``t + dt`` instead of ``t``.

    See also
    --------
    euler, rk2, rk4, milstein
    """

    # ===========================================================================
    # Parsing definitions
    # ===========================================================================
    #: Legal names for temporary variables
    TEMP_VAR = ~Literal("x_new") + Word(
        f"{string.ascii_letters}_", f"{string.ascii_letters + string.digits}_"
    ).setResultsName("identifier")

    #: A single expression
    EXPRESSION = restOfLine.setResultsName("expression")

    #: An assignment statement
    STATEMENT = Group(TEMP_VAR + Suppress("=") + EXPRESSION).setResultsName("statement")

    #: The last line of a state updater description
    OUTPUT = Group(
        Suppress(Literal("x_new")) + Suppress("=") + EXPRESSION
    ).setResultsName("output")

    #: A complete state updater description
    DESCRIPTION = ZeroOrMore(STATEMENT) + OUTPUT

    def __init__(self, description, stochastic=None, custom_check=None):
        self._description = description
        self.stochastic = stochastic
        self.custom_check = custom_check

        try:
            parsed = ExplicitStateUpdater.DESCRIPTION.parseString(
                description, parseAll=True
            )
        except ParseException as p_exc:
            ex = SyntaxError(f"Parsing failed: {str(p_exc.msg)}")
            ex.text = str(p_exc.line)
            ex.offset = p_exc.column
            ex.lineno = p_exc.lineno
            raise ex

        self.statements = []
        self.symbols = SYMBOLS.copy()
        for element in parsed:
            expression = str_to_sympy(element.expression)
            # Replace all symbols used in state updater expressions by unique
            # names that cannot clash with user-defined variables or functions
            expression = expression.subs(sympy.Function("f"), self.symbols["__f"])
            expression = expression.subs(sympy.Function("g"), self.symbols["__g"])
            symbols = list(expression.atoms(sympy.Symbol))
            unique_symbols = []
            for symbol in symbols:
                if symbol.name == "dt":
                    unique_symbols.append(symbol)
                else:
                    unique_symbols.append(_symbol(f"__{symbol.name}"))
            for symbol, unique_symbol in zip(symbols, unique_symbols):
                expression = expression.subs(symbol, unique_symbol)

            self.symbols.update({symbol.name: symbol for symbol in unique_symbols})
            if element.getName() == "statement":
                self.statements.append((f"__{element.identifier}", expression))
            elif element.getName() == "output":
                self.output = expression
            else:
                raise AssertionError(f"Unknown element name: {element.getName()}")

    def __repr__(self):
        # recreate a description string
        description = "\n".join([f"{var} = {expr}" for var, expr in self.statements])
        if len(description):
            description += "\n"
        description += f"x_new = {str(self.output)}"
        classname = self.__class__.__name__
        return f"{classname}('''{description}''', stochastic={self.stochastic!r})"

    def __str__(self):
        s = f"{self.__class__.__name__}\n"

        if len(self.statements) > 0:
            s += "Intermediate statements:\n"
            s += "\n".join(
                [f"{var} = {sympy_to_str(expr)}" for var, expr in self.statements]
            )
            s += "\n"

        s += "Output:\n"
        s += sympy_to_str(self.output)
        return s

    def _latex(self, *args):
        from sympy import Symbol, latex

        s = [r"\begin{equation}"]
        for var, expr in self.statements:
            expr = expr.subs(Symbol("x"), Symbol("x_t"))
            s.append(f"{latex(Symbol(var))} = {latex(expr)}\\\\")
        expr = self.output.subs(Symbol("x"), "x_t")
        s.append(f"x_{{t+1}} = {latex(expr)}")
        s.append(r"\end{equation}")
        return "\n".join(s)

    def _repr_latex_(self):
        return self._latex()

    def replace_func(self, x, t, expr, temp_vars, eq_symbols, stochastic_variable=None):
        """
        Used to replace a single occurance of ``f(x, t)`` or ``g(x, t)``:
        `expr` is the non-stochastic (in the case of ``f``) or stochastic
        part (``g``) of the expression defining the right-hand-side of the
        differential equation describing `var`. It replaces the variable
        `var` with the value given as `x` and `t` by the value given for
        `t`. Intermediate variables will be replaced with the appropriate
        replacements as well.

        For example, in the `rk2` integrator, the second step involves the
        calculation of ``f(k/2 + x, dt/2 + t)``.  If `var` is ``v`` and
        `expr` is ``-v / tau``, this will result in ``-(_k_v/2 + v)/tau``.

        Note that this deals with only one state variable `var`, given as
        an argument to the surrounding `_generate_RHS` function.
        """

        try:
            s_expr = str_to_sympy(str(expr))
        except SympifyError as ex:
            raise ValueError(f'Error parsing the expression "{expr}": {str(ex)}')

        for var in eq_symbols:
            # Generate specific temporary variables for the state variable,
            # e.g. '_k_v' for the state variable 'v' and the temporary
            # variable 'k'.
            if stochastic_variable is None:
                temp_var_replacements = {
                    self.symbols[temp_var]: _symbol(f"{temp_var}_{var}")
                    for temp_var in temp_vars
                }
            else:
                temp_var_replacements = {
                    self.symbols[temp_var]: _symbol(
                        f"{temp_var}_{var}_{stochastic_variable}"
                    )
                    for temp_var in temp_vars
                }
            # In the expression given as 'x', replace 'x' by the variable
            # 'var' and all the temporary variables by their
            # variable-specific counterparts.
            x_replacement = x.subs(self.symbols["__x"], eq_symbols[var])
            x_replacement = x_replacement.subs(temp_var_replacements)

            # Replace the variable `var` in the expression by the new `x`
            # expression
            s_expr = s_expr.subs(eq_symbols[var], x_replacement)

        # If the expression given for t in the state updater description
        # is not just "t" (or rather "__t"), then replace t in the
        # equations by it, and replace "__t" by "t" afterwards.
        if t != self.symbols["__t"]:
            s_expr = s_expr.subs(SYMBOLS["t"], t)
            s_expr = s_expr.replace(self.symbols["__t"], SYMBOLS["t"])

        return s_expr

    def _non_stochastic_part(
        self,
        eq_symbols,
        non_stochastic,
        non_stochastic_expr,
        stochastic_variable,
        temp_vars,
        var,
    ):
        non_stochastic_results = []
        if stochastic_variable is None or len(stochastic_variable) == 0:
            # Replace the f(x, t) part
            replace_f = lambda x, t: self.replace_func(
                x, t, non_stochastic, temp_vars, eq_symbols
            )
            non_stochastic_result = non_stochastic_expr.replace(
                self.symbols["__f"], replace_f
            )
            # Replace x by the respective variable
            non_stochastic_result = non_stochastic_result.subs(
                self.symbols["__x"], eq_symbols[var]
            )
            # Replace intermediate variables
            temp_var_replacements = {
                self.symbols[temp_var]: _symbol(f"{temp_var}_{var}")
                for temp_var in temp_vars
            }
            non_stochastic_result = non_stochastic_result.subs(temp_var_replacements)
            non_stochastic_results.append(non_stochastic_result)
        elif isinstance(stochastic_variable, str):
            # Replace the f(x, t) part
            replace_f = lambda x, t: self.replace_func(
                x, t, non_stochastic, temp_vars, eq_symbols, stochastic_variable
            )
            non_stochastic_result = non_stochastic_expr.replace(
                self.symbols["__f"], replace_f
            )
            # Replace x by the respective variable
            non_stochastic_result = non_stochastic_result.subs(
                self.symbols["__x"], eq_symbols[var]
            )
            # Replace intermediate variables
            temp_var_replacements = {
                self.symbols[temp_var]: _symbol(
                    f"{temp_var}_{var}_{stochastic_variable}"
                )
                for temp_var in temp_vars
            }

            non_stochastic_result = non_stochastic_result.subs(temp_var_replacements)
            non_stochastic_results.append(non_stochastic_result)
        else:
            # Replace the f(x, t) part
            replace_f = lambda x, t: self.replace_func(
                x, t, non_stochastic, temp_vars, eq_symbols
            )
            non_stochastic_result = non_stochastic_expr.replace(
                self.symbols["__f"], replace_f
            )
            # Replace x by the respective variable
            non_stochastic_result = non_stochastic_result.subs(
                self.symbols["__x"], eq_symbols[var]
            )
            # Replace intermediate variables
            temp_var_replacements = {
                self.symbols[temp_var]: reduce(
                    operator.add,
                    [_symbol(f"{temp_var}_{var}_{xi}") for xi in stochastic_variable],
                )
                for temp_var in temp_vars
            }

            non_stochastic_result = non_stochastic_result.subs(temp_var_replacements)
            non_stochastic_results.append(non_stochastic_result)

        return non_stochastic_results

    def _stochastic_part(
        self,
        eq_symbols,
        stochastic,
        stochastic_expr,
        stochastic_variable,
        temp_vars,
        var,
    ):
        stochastic_results = []
        if isinstance(stochastic_variable, str):
            # Replace the g(x, t) part
            replace_f = lambda x, t: self.replace_func(
                x,
                t,
                stochastic.get(stochastic_variable, 0),
                temp_vars,
                eq_symbols,
                stochastic_variable,
            )
            stochastic_result = stochastic_expr.replace(self.symbols["__g"], replace_f)
            # Replace x by the respective variable
            stochastic_result = stochastic_result.subs(
                self.symbols["__x"], eq_symbols[var]
            )
            # Replace dW by the respective variable
            stochastic_result = stochastic_result.subs(
                self.symbols["__dW"], stochastic_variable
            )
            # Replace intermediate variables
            temp_var_replacements = {
                self.symbols[temp_var]: _symbol(
                    f"{temp_var}_{var}_{stochastic_variable}"
                )
                for temp_var in temp_vars
            }

            stochastic_result = stochastic_result.subs(temp_var_replacements)
            stochastic_results.append(stochastic_result)
        else:
            for xi in stochastic_variable:
                # Replace the g(x, t) part
                replace_f = lambda x, t: self.replace_func(
                    x, t, stochastic.get(xi, 0), temp_vars, eq_symbols, xi  # noqa: B023
                )
                stochastic_result = stochastic_expr.replace(
                    self.symbols["__g"], replace_f
                )
                # Replace x by the respective variable
                stochastic_result = stochastic_result.subs(
                    self.symbols["__x"], eq_symbols[var]
                )

                # Replace dW by the respective variable
                stochastic_result = stochastic_result.subs(self.symbols["__dW"], xi)

                # Replace intermediate variables
                temp_var_replacements = {
                    self.symbols[temp_var]: _symbol(f"{temp_var}_{var}_{xi}")
                    for temp_var in temp_vars
                }

                stochastic_result = stochastic_result.subs(temp_var_replacements)
                stochastic_results.append(stochastic_result)
        return stochastic_results

    def _generate_RHS(
        self,
        eqs,
        var,
        eq_symbols,
        temp_vars,
        expr,
        non_stochastic_expr,
        stochastic_expr,
        stochastic_variable=(),
    ):
        """
        Helper function used in `__call__`. Generates the right hand side of
        an abstract code statement by appropriately replacing f, g and t.
        For example, given a differential equation ``dv/dt = -(v + I) / tau``
        (i.e. `var` is ``v` and `expr` is ``(-v + I) / tau``) together with
        the `rk2` step ``return x + dt*f(x +  k/2, t + dt/2)``
        (i.e. `non_stochastic_expr` is
        ``x + dt*f(x +  k/2, t + dt/2)`` and `stochastic_expr` is ``None``),
        produces ``v + dt*(-v - _k_v/2 + I + _k_I/2)/tau``.

        """

        # Note: in the following we are silently ignoring the case that a
        # state updater does not care about either the non-stochastic or the
        # stochastic part of an equation. We do trust state updaters to
        # correctly specify their own abilities (i.e. they do not claim to
        # support stochastic equations but actually just ignore the stochastic
        # part). We can't really check the issue here, as we are only dealing
        # with one line of the state updater description. It is perfectly valid
        # to write the euler update as:
        #     non_stochastic = dt * f(x, t)
        #     stochastic = dt**.5 * g(x, t) * xi
        #     return x + non_stochastic + stochastic
        #
        # In the above case, we'll deal with lines which do not define either
        # the stochastic or the non-stochastic part.

        non_stochastic, stochastic = expr.split_stochastic()

        if non_stochastic_expr is not None:
            # We do have a non-stochastic part in the state updater description
            non_stochastic_results = self._non_stochastic_part(
                eq_symbols,
                non_stochastic,
                non_stochastic_expr,
                stochastic_variable,
                temp_vars,
                var,
            )
        else:
            non_stochastic_results = []

        if not (stochastic is None or stochastic_expr is None):
            # We do have a stochastic part in the state
            # updater description
            stochastic_results = self._stochastic_part(
                eq_symbols,
                stochastic,
                stochastic_expr,
                stochastic_variable,
                temp_vars,
                var,
            )
        else:
            stochastic_results = []

        RHS = sympy.Number(0)
        # All the parts (one non-stochastic and potentially more than one
        # stochastic part) are combined with addition
        for non_stochastic_result in non_stochastic_results:
            RHS += non_stochastic_result
        for stochastic_result in stochastic_results:
            RHS += stochastic_result

        return sympy_to_str(RHS)

    def __call__(self, eqs, variables=None, method_options=None):
        """
        Apply a state updater description to model equations.

        Parameters
        ----------
        eqs : `Equations`
            The equations describing the model
        variables: dict-like, optional
            The `Variable` objects for the model. Ignored by the explicit
            state updater.
        method_options : dict, optional
            Additional options to the state updater (not used at the moment
            for the explicit state updaters).

        Examples
        --------
        >>> from brian2 import *
        >>> eqs = Equations('dv/dt = -v / tau : volt')
        >>> print(euler(eqs))
        _v = -dt*v/tau + v
        v = _v
        >>> print(rk4(eqs))
        __k_1_v = -dt*v/tau
        __k_2_v = -dt*(__k_1_v/2 + v)/tau
        __k_3_v = -dt*(__k_2_v/2 + v)/tau
        __k_4_v = -dt*(__k_3_v + v)/tau
        _v = __k_1_v/6 + __k_2_v/3 + __k_3_v/3 + __k_4_v/6 + v
        v = _v
        """
        extract_method_options(method_options, {})
        # Non-stochastic numerical integrators should work for all equations,
        # except for stochastic equations
        if eqs.is_stochastic and self.stochastic is None:
            raise UnsupportedEquationsException(
                "Cannot integrate stochastic equations with this state updater."
            )
        if self.custom_check:
            self.custom_check(eqs, variables)
        # The final list of statements
        statements = []

        stochastic_variables = eqs.stochastic_variables

        # The variables for the intermediate results in the state updater
        # description, e.g. the variable k in rk2
        intermediate_vars = [var for var, expr in self.statements]

        # A dictionary mapping all the variables in the equations to their
        # sympy representations
        eq_variables = {var: _symbol(var) for var in eqs.eq_names}

        # Generate the random numbers for the stochastic variables
        for stochastic_variable in stochastic_variables:
            statements.append(f"{stochastic_variable} = dt**.5 * randn()")

        substituted_expressions = eqs.get_substituted_expressions(variables)

        # Process the intermediate statements in the stateupdater description
        for intermediate_var, intermediate_expr in self.statements:
            # Split the expression into a non-stochastic and a stochastic part
            non_stochastic_expr, stochastic_expr = split_expression(intermediate_expr)

            # Execute the statement by appropriately replacing the functions f
            # and g and the variable x for every equation in the model.
            # We use the model equations where the subexpressions have
            # already been substituted into the model equations.
            for var, expr in substituted_expressions:
                for xi in stochastic_variables:
                    RHS = self._generate_RHS(
                        eqs,
                        var,
                        eq_variables,
                        intermediate_vars,
                        expr,
                        non_stochastic_expr,
                        stochastic_expr,
                        xi,
                    )
                    statements.append(f"{intermediate_var}_{var}_{xi} = {RHS}")
                if not stochastic_variables:  # no stochastic variables
                    RHS = self._generate_RHS(
                        eqs,
                        var,
                        eq_variables,
                        intermediate_vars,
                        expr,
                        non_stochastic_expr,
                        stochastic_expr,
                    )
                    statements.append(f"{intermediate_var}_{var} = {RHS}")

        # Process the "return" line of the stateupdater description
        non_stochastic_expr, stochastic_expr = split_expression(self.output)

        if eqs.is_stochastic and (
            self.stochastic != "multiplicative"
            and eqs.stochastic_type == "multiplicative"
        ):
            # The equations are marked as having multiplicative noise and the
            # current state updater does not support such equations. However,
            # it is possible that the equations do not use multiplicative noise
            # at all. They could depend on time via a function that is constant
            # over a single time step (most likely, a TimedArray). In that case
            # we can integrate the equations
            dt_value = variables["dt"].get_value()[0] if "dt" in variables else None
            for _, expr in substituted_expressions:
                _, stoch = expr.split_stochastic()
                if stoch is None:
                    continue
                # There could be more than one stochastic variable (e.g. xi_1, xi_2)
                for _, stoch_expr in stoch.items():
                    sympy_expr = str_to_sympy(stoch_expr.code)
                    # The equation really has multiplicative noise, if it depends
                    # on time (and not only via a function that is constant
                    # over dt), or if it depends on another variable defined
                    # via differential equations.
                    if not is_constant_over_dt(sympy_expr, variables, dt_value) or len(
                        stoch_expr.identifiers & eqs.diff_eq_names
                    ):
                        raise UnsupportedEquationsException(
                            "Cannot integrate "
                            "equations with "
                            "multiplicative noise with "
                            "this state updater."
                        )

        # Assign a value to all the model variables described by differential
        # equations
        for var, expr in substituted_expressions:
            RHS = self._generate_RHS(
                eqs,
                var,
                eq_variables,
                intermediate_vars,
                expr,
                non_stochastic_expr,
                stochastic_expr,
                stochastic_variables,
            )
            statements.append(f"_{var} = {RHS}")

        # Assign everything to the final variables
        for var, _ in substituted_expressions:
            statements.append(f"{var} = _{var}")

        return "\n".join(statements)


# ===============================================================================
# Excplicit state updaters
# ===============================================================================

# these objects can be used like functions because they are callable

#: Forward Euler state updater
euler = ExplicitStateUpdater(
    "x_new = x + dt * f(x,t) + g(x,t) * dW", stochastic="additive"
)

#: Second order Runge-Kutta method (midpoint method)
rk2 = ExplicitStateUpdater(
    """
    k = dt * f(x,t)
    x_new = x + dt*f(x +  k/2, t + dt/2)"""
)

#: Classical Runge-Kutta method (RK4)
rk4 = ExplicitStateUpdater(
    """
    k_1 = dt*f(x,t)
    k_2 = dt*f(x+k_1/2,t+dt/2)
    k_3 = dt*f(x+k_2/2,t+dt/2)
    k_4 = dt*f(x+k_3,t+dt)
    x_new = x+(k_1+2*k_2+2*k_3+k_4)/6
    """
)


def diagonal_noise(equations, variables):
    """
    Checks whether we deal with diagonal noise, i.e. one independent noise
    variable per variable.

    Raises
    ------
    UnsupportedEquationsException
        If the noise is not diagonal.
    """
    if not equations.is_stochastic:
        return

    stochastic_vars = []
    for _, expr in equations.get_substituted_expressions(variables):
        expr_stochastic_vars = expr.stochastic_variables
        if len(expr_stochastic_vars) > 1:
            # More than one stochastic variable --> no diagonal noise
            raise UnsupportedEquationsException(
                "Cannot integrate stochastic "
                "equations with non-diagonal "
                "noise with this state "
                "updater."
            )
        stochastic_vars.extend(expr_stochastic_vars)

    # If there's no stochastic variable is used in more than one equation, we
    # have diagonal noise
    if len(stochastic_vars) != len(set(stochastic_vars)):
        raise UnsupportedEquationsException(
            "Cannot integrate stochastic "
            "equations with non-diagonal "
            "noise with this state "
            "updater."
        )


#: Derivative-free Milstein method
milstein = ExplicitStateUpdater(
    """
    x_support = x + dt*f(x, t) + dt**.5 * g(x, t)
    g_support = g(x_support, t)
    k = 1/(2*dt**.5)*(g_support - g(x, t))*(dW**2)
    x_new = x + dt*f(x,t) + g(x, t) * dW + k
    """,
    stochastic="multiplicative",
    custom_check=diagonal_noise,
)

#: Stochastic Heun method (for multiplicative Stratonovic SDEs with non-diagonal
#: diffusion matrix)
heun = ExplicitStateUpdater(
    """
    x_support = x + g(x,t) * dW
    g_support = g(x_support,t+dt)
    x_new = x + dt*f(x,t) + .5*dW*(g(x,t)+g_support)
    """,
    stochastic="multiplicative",
)
